using System;

namespace Indexer.Engine
{
	public interface IParser
	{
		void Initialize();
		bool Accepts(string fileName, byte[] previewBuffer);
		bool Accepts(string fileName, string previewBuffer);
		string Extension { get; }
		void Copy(string fileName, System.IO.TextWriter output);
		bool SupportsCopy { get; }
	}
}

namespace Indexer.Engine.Parsers
{
	public abstract class ParserBase: IParser
	{
		private string _extension;
		
		public ParserBase(string extension)
		{
			_extension = extension;
		}
		
		public string Extension 
		{
			get 
			{
				return _extension;
			}
		}
		
		protected static bool BufferStartsWith(string buffer, string pattern)
		{
			return buffer.StartsWith(pattern);
		}
		
		protected static bool BufferStartsWith(byte[] buffer, string pattern)
		{
			if ((buffer == null) || (buffer.Length < pattern.Length))
			{
				return false;
			}
			for(int i = 0; i < pattern.Length; i++)
			{
				if (buffer[i] != Convert.ToByte(pattern[i]))
				{
					return false;
				}
			}
			return true;
		}
		
		protected static bool BufferStartsWith(byte[] buffer, byte[] pattern)
		{
			if ((buffer == null) || (buffer.Length < pattern.Length))
			{
				return false;
			}
			for(int i = 0; i < pattern.Length; i++)
			{
				if (buffer[i] != pattern[i])
				{
					return false;
				}
			}
			return true;
		}
		
		protected static string BufferToString(byte[] buffer, bool uniformCasing)
		{
			System.Text.StringBuilder result = new System.Text.StringBuilder(buffer.Length);
			for(int i = 0; i < buffer.Length; i++)
			{
				char c = (char)buffer[i];
				if (uniformCasing)
				{
					c = char.ToLower(c);
				}
				result.Append(c);
			}
			return result.ToString();
		}
		
		protected static string BufferToString(string buffer, bool uniformCasing)
		{
			if (uniformCasing)
			{
				return buffer.ToLower();
			}
			else
			{
				return buffer;
			}
		}
		
		protected int IndexOf(byte[] buffer, byte[] pattern, bool caseSensitive)
		{
			if ((buffer == null) || (buffer.Length < pattern.Length))
			{
				return -1;
			}
			bool uniformCasing = !caseSensitive;
			return BufferToString(buffer, uniformCasing).IndexOf(BufferToString(pattern, uniformCasing));
		}
		
		protected int IndexOf(byte[] buffer, string pattern, bool caseSensitive)
		{
			if ((buffer == null) || (buffer.Length < pattern.Length))
			{
				return -1;
			}
			bool uniformCasing = !caseSensitive;
			return BufferToString(buffer, uniformCasing).IndexOf(BufferToString(pattern, uniformCasing));
		}
		
		protected int IndexOf(string buffer, string pattern, bool caseSensitive)
		{
			if ((buffer == null) || (buffer.Length < pattern.Length))
			{
				return -1;
			}
			bool uniformCasing = !caseSensitive;
			return BufferToString(buffer, uniformCasing).IndexOf(BufferToString(pattern, uniformCasing));
		}
		
		protected bool BufferContains(byte[] buffer, byte[] pattern, bool caseSensitive)
		{
			return IndexOf(buffer, pattern, caseSensitive) >= 0;
		}

		protected bool BufferContains(byte[] buffer, string pattern, bool caseSensitive)
		{
			return IndexOf(buffer, pattern, caseSensitive) >= 0;
		}
		
		protected bool BufferContains(string buffer, string pattern, bool caseSensitive)
		{
			return IndexOf(buffer, pattern, caseSensitive) >= 0;
		}
		
		protected static bool FileHasExtension(string fileName, string extension)
		{
			return FileHasExtension(fileName, new string[]{extension});
		}
		
		protected static bool FileHasExtension(string fileName, string[] extensions)
		{
			string extension = System.IO.Path.GetExtension(fileName);
			foreach(string pattern in extensions)
			{
				if (String.Compare(extension, pattern, true) == 0)
				{
					return true;
				}
			}
			return false;
		}
		
		public abstract bool Accepts(string fileName, byte[] previewBuffer);
		public abstract bool Accepts(string fileName, string previewBuffer);
		
		public virtual void Copy(string fileName, System.IO.TextWriter output)
		{
			throw new NotSupportedException();
		}
		
		protected virtual bool GetSupportsCopy()
		{
			return false;
		}
		
		public bool SupportsCopy
		{
			get
			{
				return GetSupportsCopy();
			}
		}
		
		public virtual void Initialize()
		{
		}
	}
	
	/// <summary>
	/// Purpose of this class is to capture commands from parsers and translate them
	/// into "readable" text used later on in searching and showing search hits
	/// </summary>
	internal class Writer
	{
		private System.IO.TextWriter _stream;
		private System.Text.StringBuilder _buffer = new System.Text.StringBuilder();
		private int _numberOfEmptyLinesSent = 1;
		
		private bool FlushLineBuffer()
		{
			FlushCharacterBuffer(true);
			
			bool sentSomeData = false;
			
			if (_buffer.Length > 0)
			{
				string line = _buffer.ToString();
				bool thisLineIsEmpty = String.IsNullOrEmpty(line.Trim());
				
				_buffer = new System.Text.StringBuilder();
				
				if (!thisLineIsEmpty)
				{
					_stream.Write(line);
					sentSomeData = true;
				}
			}
			return sentSomeData;
		}
		
		public Writer(System.IO.TextWriter stream)
		{
			_stream = stream;
		}
		
		private void StartNewLine()
		{
			if (FlushLineBuffer())
			{
				_numberOfEmptyLinesSent = 0;
				_stream.WriteLine();
			}
			else
			{
				if (_numberOfEmptyLinesSent == 0)
				{
					_stream.WriteLine();
					_numberOfEmptyLinesSent++;
				}
			}
			_lastCharacterWasWhiteSpace = true;
		}
		
		private void WriteCharacter(string value)
		{
			if (!String.IsNullOrEmpty(value))
			{
				for(int i = 0; i < value.Length; i++)
				{
					WriteCharacter(value[i]);
				}
			}
		}
		
		#region Character writer
		private bool _lastCharacterWasWhiteSpace = true;
		private char _toBeFlushedCharacter;
		private bool _somethingToBeFlushed = false;
		
		private void FlushCharacterBuffer(bool skipWhiteSpace)
		{
			if (_somethingToBeFlushed)
			{
				_somethingToBeFlushed = false;
				if (!(skipWhiteSpace && _lastCharacterWasWhiteSpace))
				{
					switch (_toBeFlushedCharacter)
					{
						case '\t':
							_buffer.Append(", ");
							break;
							
						default:
							_buffer.Append(_toBeFlushedCharacter);
							break;
					}
				}
			}
		}
		
		private void WriteCharacter(char value)
		{
			bool thisCharacterIsBlank = (value == ' ') || (value == '\t');
			if (thisCharacterIsBlank && _lastCharacterWasWhiteSpace)
			{
				// use this character instead of the one waiting in emitor
				// if this one is the "\t" -> of all consequent white spaces
				// if there is a '\t' then it will be output
				if ((value == '\t') && _somethingToBeFlushed)
				{
					_toBeFlushedCharacter = value;
				}
			}
			else
			{
				FlushCharacterBuffer(false);
				_somethingToBeFlushed = true;
				_toBeFlushedCharacter = value;
				_lastCharacterWasWhiteSpace = thisCharacterIsBlank;
			}
		}
		#endregion
		
		public void Write(string buffer)
		{
			if (!String.IsNullOrEmpty(buffer))
			{
				for(int i = 0; i < buffer.Length; i++)
				{
					Write(buffer[i]);
				}
			}
		}
		
		private void Write(char c)
		{
			switch (c)
			{
				case '\r':
					break;
					
				case '\n':
					StartNewLine();
					break;
					
				case '\u201C':	// left upper quote
				case '\u201D':	// right upper quote
				case '\u201E':	// left lower quote
					WriteCharacter("\"");
					break;
				
				case '\u0060':
				case '\u2018':
				case '\u2019':
				case '\u201B':
					WriteCharacter("'");
					break;
								
				case '\u2013':
					WriteCharacter("-");
					break;
								
				case '\u2026':
					WriteCharacter("...");
					break;
								
				case '\u00A0':
					WriteCharacter(" ");
					break;
					
				case '\u00B7':
				case '\u2022':
					WriteCharacter("* ");
					break;

				case '\u000B':			// paragraph break?
					StartNewLine();
					break;
								
				case '\u000C':			// page break
					StartNewLine();
					break;
					
				default:
					WriteCharacter(c);
					break;
			}
		}
		
		public void Write(char[] buffer, int count)
		{
			for(int i = 0; i < count; i++)
			{
				Write(buffer[i]);
			}
		}
		
		public void WriteLine()
		{
			Write(new char[]{'\r', '\n'}, 2);
		}
		
		public void WriteLine(string line)
		{
			Write(line);
			WriteLine();
		}
		
		public void Close()
		{
			if (_stream != null)
			{
				FlushLineBuffer();
				_stream = null;
			}
		}
	}

	public class Parsers
	{
		private static System.Collections.Generic.List<IParser> _parsers = new System.Collections.Generic.List<IParser>();
		private static System.Collections.Generic.List<IParser> _detectors = new System.Collections.Generic.List<IParser>();
		
		static Parsers()
		{
			// binary parsers
			_parsers.Add(new PdfParser());
			_parsers.Add(new OfficeParser(".doc"));
			_parsers.Add(new OfficeParser(".rtf", ".doc"));	// some files named *.rtf actually contain *.doc content
			_parsers.Add(new OfficeParser(".xls"));
			_parsers.Add(new OfficeParser(".ppt"));
			// TODO: this does not work - _parsers.Add(new OfficeParser(".msg", ".eml"));	// Outlook message file
			_parsers.Add(new RtfParser());
			_parsers.Add(new HtmlParser());
			_parsers.Add(new MailParser());
			// TODO: add other office extensions
			
			// text detectors
			_detectors.Add(new XmlDetector());
		}
		
		private static bool _initialized = false;
		private static void Initialize()
		{
			if (!_initialized)
			{
				_initialized = true;
				foreach(IParser item in _parsers)
				{
					item.Initialize();
				}
				foreach(IParser item in _detectors)
				{
					item.Initialize();
				}
			}
		}
		
		public static IParser Match(string fileName, byte[] previewBuffer)
		{
			Initialize();
			
			foreach(IParser result in _parsers)
			{
				if (result.Accepts(fileName, previewBuffer))
				{
					return result;
				}
			}
			foreach(IParser result in _detectors)
			{
				if (result.Accepts(fileName, previewBuffer))
				{
					return result;
				}
			}
			return null;
		}
		
		public static IParser Match(string fileName, string previewBuffer)
		{
			Initialize();
			
			foreach(IParser result in _parsers)
			{
				if (result.Accepts(fileName, previewBuffer))
				{
					return result;
				}
			}
			return null;
		}
	}
}
