using System;
using Indexer.Engine;
using DocumentHandle = System.Int32;

namespace Indexer.Gui
{
	public class SearchCommand: WorkItem
	{
		public Index Index;
		public Index.Query Query;
		public Index.DocumentCollection Result;
		
		public override void Execute()
		{
			Result = Index.Execute(Query);
		}
		
		protected override void DisconnectThread()
		{
			Index.IO.DisconnectThread();
		}
	}
	
	public class OpenIndexCommand: WorkItem
	{
		public Index Result;
		public string FileName;
		public IIndexObserver Observer;
		
		public override void Execute()
		{
			if (String.IsNullOrEmpty(FileName))
			{
				Result = new Index(Observer);
			}
			else
			{
				Result = new Index(Observer, FileName);
			}
			try
			{
				try
				{
					// TODO: use write message?
					Result.Prepare();
				}
				catch(Exception)
				{
					Result.Initialized(false);
					throw;
				}
				Result.Initialized(true);
			}
			catch(Exception)
			{
				Result.Dispose();
				Result = null;
				throw;
			}
		}
		
		protected override void DisconnectThread()
		{
			if (Result != null)
			{
				Result.IO.DisconnectThread();
			}
		}
	}
	
	public class SaveIndexCommand: WorkItem
	{
		public Index Index;
		public string FileName;
		
		public override void Execute()
		{
			if (String.IsNullOrEmpty(FileName))
			{
				Index.Save();
			}
			else
			{
				Index.SaveAs(FileName);
			}
		}
		
		protected override void DisconnectThread()
		{
			Index.IO.DisconnectThread();
		}
	}
	
	public class RebuildIndexCommand: WorkItem
	{
		public Index Index;
		
		public override void Execute()
		{
			Index.Rebuild();
		}
		
		protected override void DisconnectThread()
		{
			Index.IO.DisconnectThread();
		}
	}
	
	public class UpdateIndexCommand: WorkItem
	{
		public Index Index;
		public bool SkipCheckingPhase = false;
		
		public override void Execute()
		{
			Index.Update(SkipCheckingPhase);
		}
		
		protected override void DisconnectThread()
		{
			Index.IO.DisconnectThread();
		}
	}
	
	public class AddToIndexCommand: WorkItem
	{
		public Index Index;
		public string[] FileNames;
		
		public override void Execute()
		{
			Index.Add(FileNames);
		}
		
		protected override void DisconnectThread()
		{
			Index.IO.DisconnectThread();
		}
	}

	public class MountArchiveCommand: WorkItem
	{
		public Index Index;
		public string[] FileNames;
		
		public override void Execute()
		{
			Index.Mount(FileNames);
		}
		
		protected override void DisconnectThread()
		{
			Index.IO.DisconnectThread();
		}
	}
	
#if false
	// TODO: bookmarkJumps don't work
	public class ExportFilesCommand: WorkItem
	{
		public string[] FileNames;
		public System.Collections.BitArray Colorize;
		public System.Collections.Generic.List<string> TemporaryFiles;
		public Index Index;
		public Index.Query Query;
		public string Title;
		public string Result;
		
		private string GetTempFile(string extension)
		{
			string tempFile = System.IO.Path.GetTempFileName();
			string result = tempFile + extension;
			System.IO.File.Create(result).Dispose();
			System.IO.File.Delete(tempFile);
			TemporaryFiles.Add(result);
			return result;
		}
		
		public override void Execute()
		{
			System.Threading.SlicedProgress progress = new System.Threading.SlicedProgress(Progress);
			
			Result = GetTempFile(".mht");
			System.Web.ArchiveWriter output = new System.Web.ArchiveWriter(Result);
			
			string masterFile = output.CreateID(".htm");
			System.IO.StringWriter contents = new System.IO.StringWriter();
			string firstIndexFile = null;
			
			for(int i = 0; i < FileNames.Length; i++)
			{
				progress.Add(100);
			}
			try
			{
				contents.WriteLine("<html><head><meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\"></head><body style='color:#000000; background:#ffffff; '><pre>");
				
				int toGo = FileNames.Length;
				int index = -1;
				foreach(string fileName in FileNames)
				{
					index++;
					toGo--;
					
					if (Colorize[index])
					{
						string tempFile = output.CreateID(".htm");
						Highlighter.Strings previews = new Highlighter.Strings();
						Highlighter.Strings bookmarks;
						System.IO.TextWriter w = output.CreatePage(tempFile);
						try
						{
							bookmarks = Highlighter.FormatDocument(fileName, Index, Query, w, previews, progress[index]);
						}
						finally
						{
							w.Dispose();
						}
						if ((bookmarks != null) && (bookmarks.Count > 0))
						{
							string[] bookmarkJumps = new string[bookmarks.Count];
							for(int i = 0; i < bookmarks.Count; i++)
							{
								bookmarkJumps[i] = output.CreateID(".htm");
								//output.AddPage(bookmarkJumps[i], String.Format("<html><head><meta http-equiv=\"Refresh\" content=\"0; url={0};\"></head></html>", System.IO.Path.GetFileName(tempFile), bookmarks[i]));
								output.AddPage(bookmarkJumps[i], String.Format("<html><body><script type=text/javascript>window.location=\"{0}#{1}\";</script></body></html>", System.IO.Path.GetFileName(tempFile), bookmarks[i]));
							}
							
							if (String.IsNullOrEmpty(firstIndexFile))
							{
								firstIndexFile = bookmarkJumps[0];
							}
							
							contents.Write(String.Format("<a href=\"{0}\" target=main><b>{1}</b></a><ol>", System.IO.Path.GetFileName(tempFile), Highlighter.HtmlEncode(fileName)));
									
							for(int i = 0; i < bookmarks.Count; i++)
							{
								string suffix = "";
								if (i == bookmarks.Count - 1)
								{
									suffix = "\r\n</ol>";	
								}
								contents.WriteLine(String.Format("<li><a href=\"{0}\" target=main>{1}</a>" + suffix, bookmarkJumps[i], previews[i]));
							}
							/*
							if (toGo > 0)
							{
								contents.WriteLine("<hr size=1 noshade>");
							}
							*/
						}
						else
						{
							if (String.IsNullOrEmpty(firstIndexFile))
							{
								firstIndexFile = System.IO.Path.GetFileName(tempFile);
							}
							contents.WriteLine(String.Format("<a href=\"{0}\" target=main><b>{1}</b></a>", System.IO.Path.GetFileName(tempFile), Highlighter.HtmlEncode(fileName)));
						}
					}
					else
					{
						contents.WriteLine(String.Format("<span style='color:#808080; '><b>{0}</b></span>", Highlighter.HtmlEncode(fileName)));
					}
				}
				contents.WriteLine("</pre></body></html>");
			}
			finally
			{
				progress.Close();
				
				if (contents != null)
				{
					output.AddPage(masterFile, contents.ToString());
						
					contents.Dispose();
					contents = null;
				}
			}
			
			{
				System.IO.TextWriter w = output.CreatePage("frameset.htm", true);
				try
				{
					w.Write(String.Format("<html><head><title>{0}</title></head>", Highlighter.HtmlEncode(Title)));
					w.Write(String.Format("<frameset rows=\"20%,*\"><frame src=\"{0}\"><frame src=\"{1}\" name=main></frameset>", masterFile, firstIndexFile));
					w.Write("</html>");
				}
				finally
				{
					w.Dispose();
				}
			}
			
			output.Close();
		}
	}
#else
	public class ExportFilesCommand: WorkItem, Highlighter.IMatchMarker
	{
		public string[] FileNames;
		public DocumentHandle[] Handles;
		public System.Collections.BitArray Colorize;
		public System.Collections.Generic.List<string> TemporaryFiles;
		public Index Index;
		public Index.Query Query;
		public string Title;
		public string Result;
		
		private string GetTempHtmlFile()
		{
			string tempFile = System.IO.Path.GetTempFileName();
			string result = tempFile + ".htm";
			System.IO.File.Create(result).Dispose();
			System.IO.File.Delete(tempFile);
			TemporaryFiles.Add(result);
			return result;
		}
		
		public override void Execute()
		{
			System.Threading.SlicedProgress progress = new System.Threading.SlicedProgress(Progress);
			
			string masterFile = GetTempHtmlFile();
			System.IO.TextWriter contents = new System.IO.StreamWriter(masterFile);
			string firstIndexFile = null;
			
			for(int i = 0; i < FileNames.Length; i++)
			{
				progress.Add(100);
			}
			try
			{
				contents.WriteLine("<html><head><meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\"></head><body style='color:#000000; background:#ffffff; '><pre>");
				
				int toGo = FileNames.Length;
				int index = -1;
				foreach(string fileName in FileNames)
				{
					index++;
					toGo--;
					
					if (Colorize[index])
					{
						string tempFile = GetTempHtmlFile();
						Highlighter.Strings previews = new Highlighter.Strings();
						Highlighter.Strings bookmarks;
						System.IO.TextReader sourceStream = this.Index.CreateTextReader(Handles[index]);
						Highlighter.Options options = new Highlighter.Options();
						if (this.Index.Documents.GetKind(Handles[index]) == (Index.DocumentKind.Indexed | Index.DocumentKind.Cached))
						{
							options.Extension = ".txt";
							options.WordWrap = true;
							// options.DimmedLineNumbers = true;
							options.UseLineNumbers = false;
							options.Colorize = false;
							options.PreserveSpaces = false;
						}
						else
						{
							// TODO: ?
							options.WordWrap = true;
						}
						try
						{
							bookmarks = Highlighter.FormatDocument(sourceStream, fileName, options, this, tempFile, previews, progress[index]);
						}
						finally
						{
							sourceStream.Dispose();
						}
						if ((bookmarks != null) && (bookmarks.Count > 0))
						{
							if (String.IsNullOrEmpty(firstIndexFile))
							{
								firstIndexFile = System.IO.Path.GetFileName(tempFile) + "#" + bookmarks[0];
							}
							
							contents.WriteLine(String.Format("<a href=\"{0}\" target=main><b>{1}</b></a>", System.IO.Path.GetFileName(tempFile), Highlighter.HtmlEncode(fileName)));
									
							for(int i = 0; i < bookmarks.Count; i++)
							{
								string suffix = "";
								if (i == bookmarks.Count - 1)
								{
									// TODO: ??
									// suffix = "\r\n";
								}
								contents.WriteLine(String.Format(" > <a href=\"{0}#{1}\" target=main>{2}</a>" + suffix, System.IO.Path.GetFileName(tempFile), bookmarks[i], previews[i]));
							}
							/*
							if (toGo > 0)
							{
								contents.WriteLine("<hr size=1 noshade>");
							}
							*/
						}
						else
						{
							if (String.IsNullOrEmpty(firstIndexFile))
							{
								firstIndexFile = System.IO.Path.GetFileName(tempFile);
							}
							contents.WriteLine(String.Format("<a href=\"{0}\" target=main><b>{1}</b></a>", System.IO.Path.GetFileName(tempFile), Highlighter.HtmlEncode(fileName)));
						}
					}
					else
					{
						contents.WriteLine(String.Format("<span style='color:#808080; '><b>{0}</b></span>", Highlighter.HtmlEncode(fileName)));
					}
				}
				contents.WriteLine("</pre></body></html>");
			}
			finally
			{
				progress.Close();
				
				if (contents != null)
				{
					contents.Dispose();
					contents = null;
				}
			}
			
			string framesetFile = GetTempHtmlFile();
			System.IO.TextWriter w = new System.IO.StreamWriter(framesetFile);
			try
			{
				w.Write(String.Format("<html><head><title>{0}</title></head>", Highlighter.HtmlEncode(Title)));
				w.Write(String.Format("<frameset rows=\"20%,*\"><frame src=\"{0}\"><frame src=\"{1}\" name=main></frameset>", System.IO.Path.GetFileName(masterFile), firstIndexFile));
				w.Write("</html>");
			}
			finally
			{
				w.Dispose();
			}
			
			Result = framesetFile;
		}
		
		public string[] Phrases 
		{
			get 
			{
				return Query.Phrases;
			}
		}
		
		public Indexer.Engine.Highlighter.Match[] GetHighlightPoints(string text)
		{
			Indexer.Engine.Index.SearchResult[] matches = Index.Execute(Query, text);
			if (matches == null)
			{
				return null;
			}
			else
			{
				Indexer.Engine.Highlighter.Match[] results = new Indexer.Engine.Highlighter.Match[matches.Length];
				for(int i = 0; i < matches.Length; i++)
				{
					Indexer.Engine.Index.SearchResult a = matches[i];
					Indexer.Engine.Highlighter.Match b = new Indexer.Engine.Highlighter.Match();
					b.Row = a.Row;
					b.Column = a.Column;
					b.Length = a.Length;
					results[i] = b;
				}
				return results;
			}
		}
		
		protected override void DisconnectThread()
		{
			Index.IO.DisconnectThread();
		}
	}
#endif
	
	public class OpenFileCommand: WorkItem
	{
		public string FileName;
		public string Arguments;
		public Explorer.Model.Reader IO;
		public System.Collections.Generic.List<string> TemporaryFiles;
		
		public enum OpenMode
		{
			Open,
			Ask,
			OpenOrAsk
		}
		public OpenMode Mode = OpenMode.Open;
		
		private string CreateTemporaryFolder()
		{
			while(true)
			{
				try
				{
					string tempFile = System.IO.Path.GetTempFileName();
					System.IO.File.Delete(tempFile);
					System.IO.Directory.CreateDirectory(tempFile);
					TemporaryFiles.Add(tempFile);
					return tempFile;
				}
				catch(Exception)
				{
				}
			}
		}

		protected string CreateFileMapping(string fileName)
		{
			string folder = null;
			return CreateFileMapping(fileName, ref folder);
		}
		
		protected string CreateFileMapping(string fileName, ref string folder)
		{
			if ((IO == null) || (TemporaryFiles == null))
			{
				return fileName;
			}
			if (IO.IsFileNative(fileName))
			{
				return fileName;
			}

			string result;
			try
			{
				Progress.Write("Export");
				if (String.IsNullOrEmpty(folder))
				{
					folder = CreateTemporaryFolder();
				}
				result = System.IO.Path.Combine(folder, System.IO.Path.GetFileName(fileName));
				IO.CopyFile(fileName, result, Progress);
			}
			finally
			{
				Progress.Write(null);
			}
			return result;
		}
		
		public override void Execute()
		{
			FileName = CreateFileMapping(FileName);
			
			if (Mode == OpenMode.Ask)
			{
				System.Windows.Forms.ProcessHelper.OpenAs(FileName);
			}
			else
			{
				try
				{
					System.Diagnostics.Process p = new System.Diagnostics.Process();
					p.StartInfo.FileName = FileName;
					if (!String.IsNullOrEmpty(Arguments))
					{
						p.StartInfo.Arguments = Arguments;
					}
					p.StartInfo.WindowStyle = System.Diagnostics.ProcessWindowStyle.Maximized;
					p.Start();
					p.Dispose();
				}
				catch(Exception)
				{
					if (Mode == OpenMode.OpenOrAsk)
					{
						System.Windows.Forms.ProcessHelper.OpenAs(FileName);
					}
					else
					{
						throw;
					}
				}
			}
		}
		
		protected override void DisconnectThread()
		{
			if (IO != null)
			{
				IO.DisconnectThread();
			}
		}
	}
	
	public class CompareFilesCommand: OpenFileCommand
	{
		public string LeftArgument;
		public string RightArgument;
		
		public override void Execute()
		{
			Arguments = Arguments
				.Replace("%1", CreateFileMapping(LeftArgument))
				.Replace("%2", CreateFileMapping(RightArgument))
				.Replace("%3", LeftArgument)
				.Replace("%4", RightArgument);
			
			// disconnect temporary file cache to disable file mapping in base class
			TemporaryFiles = null;
			base.Execute();
		}
	}
	
	public class OpenContainingFolderCommand: OpenFileCommand
	{
		public string[] FileNames;
		
		public override void Execute()
		{
			string fileName;
			if (FileNames == null)
			{
				fileName = this.FileName;
			}
			else if (FileNames.Length == 1)
			{
				fileName = FileNames[0];
			}
			else
			{
				fileName = null;
			}
			
			if (!String.IsNullOrEmpty(fileName))
			{
				// open folder and select the file
				fileName = CreateFileMapping(fileName);
				
				FileName = "Explorer";
				Arguments = String.Format("/n,/select,\"{1}\"", System.IO.Path.GetDirectoryName(fileName), fileName);
			}
			else
			{
				string temporaryFolder = null;
				
				for(int i = 0; i < FileNames.Length; i++)
				{
					fileName = CreateFileMapping(FileNames[i], ref temporaryFolder);
				}
				
				FileName = "Explorer";
				Arguments = String.Format("/n,\"{0}\"", System.IO.Path.GetDirectoryName(fileName), fileName);
			}
			
			// disconnect temporary file cache to disable file mapping in base class
			TemporaryFiles = null;
			base.Execute();
		}
	}
	
	public class Model
	{
		public Model()
		{
		}
	}
}
