using System;

namespace Explorer.Model
{
	public enum EntryKind
	{
		UpFolder,
		Folder,
		File
	}
	
	public class Tools
	{
		public static bool DateTimeValid(DateTime date)
		{
			return date.Year != 1;
		}
		
		public static bool IsAbsolutePath(string path)
		{
			if (String.IsNullOrEmpty(path))
			{
				return false;
			}
			return path.StartsWith("\\") || (path.IndexOf(":\\") == 1);
		}
		
		public static int NameCompare(Entry x, Entry y)
		{
			int result = Convert.ToInt32(x.Kind) - Convert.ToInt32(y.Kind);
			if (result == 0)
			{
				return String.Compare(x.Name, y.Name, true);
			}
			else
			{
				return result;
			}
		}
		
		/* Sorts entries so that back link comes first, then folders, then files.
		 * Among same kinds, items are sorted so that the first one is preferred (if case of same name)
		 */
		public static int DateCompare(Entry x, Entry y)
		{
			int result = Convert.ToInt32(x.Kind) - Convert.ToInt32(y.Kind);
			if (result == 0)
			{
				bool xValid = DateTimeValid(x.LastWriteTime);
				bool yValid = DateTimeValid(y.LastWriteTime);
				
				if (xValid && yValid)
				{
					// TODO: reverse ?
					return -DateTime.Compare(x.LastWriteTime, y.LastWriteTime);
				}
				else if (xValid)
				{
					return 1;
				}
				else if (yValid)
				{
					return -1;
				}
				else
				{
					return 0;
				}
			}
			else
			{
				return result;
			}
		}
		
		public static string FolderNameToFileName(string path)
		{
			if (String.IsNullOrEmpty(path))
			{
				return "";
			}
			else if (path.EndsWith("\\") && !path.EndsWith(":\\"))
			{
				return path.Substring(0, path.Length - 1);
			}
			else
			{
				return path;
			}
		}
		
		public static string FileNameToFolderName(string path)
		{
			if (String.IsNullOrEmpty(path))
			{
				return "";
			}
			else if (!path.EndsWith("\\"))
			{
				return path + "\\";
			}
			else
			{
				return path;
			}
		}
		
		public static void Split(string fileName, out string path, out string name)
		{
			if (fileName.EndsWith(":\\"))
			{
				path = "";
				name = fileName;
			}
			else
			{
				int i = fileName.LastIndexOf('\\');
				if (i < 0)
				{
					path = "";
					name = fileName;
				}
				else
				{
					path = fileName.Substring(0, i + 1);
					name = fileName.Substring(i + 1);
				}
			}
		}
	}
	
	public class Entry: IDisposable
	{
		private string _path;
		public string Name;
		public long Size;
		public DateTime LastWriteTime;
		public EntryKind Kind;
		public int Index;
		public System.IO.FileAttributes Attributes;
		
		public object InternalHandle;
		
		public Entry()
		{
			Name = "";
			_path = "";
		}
		
		public void Dispose()
		{
			InternalHandle = null;
			// TODO: anything else?
		}
		
		/// <summary>
		/// slash-terminated path or empty string
		/// </summary>
		public string Path
		{
			get
			{
				return _path;
			}
			set
			{
				if (String.IsNullOrEmpty(value))
				{
					_path = "";
				}
				else
				{
					if (!value.EndsWith("\\"))
					{
						_path = value + "\\";
					}
					else
					{
						_path = value;
					}
				}
			}
		}
		
		public string FullName
		{
			get
			{
				return Path + Name;
			}
			set
			{
				string path;
				string name;
				Tools.Split(value, out path, out name);
				this.Path = path;
				this.Name = name;
			}
		}
	}
	
	public class Entries: System.Collections.Generic.List<Entry>, IDisposable
	{
		public new void Add(Entry item)
		{
			item.Index = this.Count;
			base.Add(item);
		}
		
		public void Dispose()
		{
			foreach(Entry item in this)
			{
				item.Dispose();
			}
			Clear();
			// TODO: anything else?
		}
	}
}
