using System;

namespace Prolog.Memory
{
	public class Database
	{
		public class SpecialPredicates
		{
			public const string Dot = ".";
			public const string DotId = Dot + "/2";
			public const string Clause = ":-";
			public const string ClauseId = Clause + "/2";
		}
		
		private Operators _operators = new Operators();
		
		public Operators Operators
		{
			get
			{
				return _operators;
			}
		}
		
		private System.Collections.Generic.List<ICell> _items = new System.Collections.Generic.List<ICell>();
		
		public Database()
		{
			Operators.Add(1200, OperatorType.xfx, new string[]{":-", "-->"});
			Operators.Add(1200, OperatorType.fx,  new string[]{"?-", ":-"});
			Operators.Add(1150, OperatorType.xfy, new string[]{"->"});
			Operators.Add(1100, OperatorType.fx,  new string[]{"dynamic"});
			Operators.Add(1100, OperatorType.xfy, new string[]{";"});
			Operators.Add(1000, OperatorType.xfy, new string[]{", "});
			Operators.Add( 900, OperatorType.fy,  new string[]{"\\+", "not"});
			Operators.Add( 700, OperatorType.xfx, new string[]{"=", "\\=", "is", "=..", "==", "\\==", "=:=", "=\\=", "<", ">", "=<", ">=", "@<", "@>", "@=<", "@>="});
			Operators.Add( 500, OperatorType.yfx, new string[]{"+", "-", "/\\", "\\/"});
			Operators.Add( 500, OperatorType.fx,  new string[]{"+", "-"});
			Operators.Add( 400, OperatorType.yfx, new string[]{"/", "*", ">>", "<<", ":"});
			Operators.Add( 300, OperatorType.xfx, new string[]{"mod"});
			Operators.Add( 200, OperatorType.xfy, new string[]{"^"});
		}
		
		public int Length
		{
			get
			{
				return _items.Count;
			}
		}
		
		public ICell this[int index]
		{
			get
			{
				return _items[index];
			}
		}
		
		public void Insert(ICell item)
		{
			_items.Insert(0, item);
		}
		
		public void Append(ICell item)
		{
			_items.Add(item);
		}
		
		private const string SpecialSymbols = "@#$%^&></+*-=\\~`:?;,";
		
		[Flags]
		private enum NameSymbols
		{
			None = 0,
			Blank = 1 << 0,
			Special = 1 << 1,
			Underscore = 1 << 2,
			UpperAlpha = 1 << 3,
			LowerAlpha = 1 << 4,
			Number = 1 << 5,
			Other = 1 << 6
		}
		
		private bool GetSymbols(char c, string list)
		{
			if (list.IndexOf(c) >= 0)
			{
				return true;
			}
			else
			{
				return false;
			}
		}
		
		private NameSymbols GetSymbols(string name)
		{
			NameSymbols result = NameSymbols.None;
			foreach(char c in name)
			{
				switch (c)
				{
					case '_':
						result = result | NameSymbols.Underscore;
						break;
						
					case ' ':
						result = result | NameSymbols.Blank;
						break;
						
					default:
						if (GetSymbols(c, SpecialSymbols))
						{
							result = result | NameSymbols.Special;
						}
						else if (GetSymbols(c, "0123456789"))
						{
							result = result | NameSymbols.Number;
						}
						else if (GetSymbols(c, "ABCDEFGHIJKLMNOPQRSTUVWXYZ"))
						{
							result = result | NameSymbols.UpperAlpha;
						}
						else if (GetSymbols(c, "abcdefghijklmnopqrstuvwxyz"))
						{
							result = result | NameSymbols.LowerAlpha;
						}
						else
						{
							result = result | NameSymbols.Other;
						}
						break;
				}
			}
			return result;
		}
		
		private bool IsVariableName(string name)
		{
			NameSymbols symbols = GetSymbols(name.Substring(0, 1));
			return (symbols & (NameSymbols.Underscore | NameSymbols.UpperAlpha)) != 0;
		}
		
		private bool NeedsQuoting(string name)
		{
			if (IsVariableName(name))
			{
				return true;
			}
			NameSymbols symbols = GetSymbols(name);
			if (symbols == NameSymbols.Special)
			{
				return false;
			}
			if ((symbols & NameSymbols.Blank) != 0)
			{
				return true;
			}
			if ((symbols & NameSymbols.Special) != 0)
			{
				return true;
			}
			return false;
		}
		
		private string NameToString(string name)
		{
			if (NeedsQuoting(name))
			{
				return "'" + name + "'";
			}
			else
			{
				return name;
			}
		}
		
		private bool IsListCell(ICell head)
		{
			if (head == null)
			{
				return true;
			}
			else
			{
				return head.FunctorId == SpecialPredicates.DotId;
			}
		}
		
		private string ListToString(ICell head)
		{
			System.Text.StringBuilder buffer = new System.Text.StringBuilder();
			buffer.Append("[");
			while(true)
			{
				buffer.Append(CellToString(head[0]));
				head = head[1];
				if (head == null)
				{
					break;
				}
				else
				{
					if (IsListCell(head))
					{
						buffer.Append(",");
					}
					else
					{
						buffer.Append("|");
						buffer.Append(CellToString(head));
						break;
					}
				}
			}
			buffer.Append("]");
			return buffer.ToString();
		}
		
		private string CellToString(ICell item)
		{
			if (item == null)
			{
				return "[]";
			}
			else
			{
				if (item.Arity == 0)
				{
					return NameToString(item.Name);
				}
				else if (IsListCell(item))
				{
					return ListToString(item);
				}
				else if ((item.Arity == 2) && Operators.IsBinary(item.Name))
				{
					System.Text.StringBuilder buffer = new System.Text.StringBuilder();
					buffer.Append(CellToString(item[0]));
					buffer.Append(" ");
					buffer.Append(NameToString(item.Name));
					buffer.Append(" ");
					buffer.Append(CellToString(item[1]));
					return buffer.ToString();
				}
				else
				{
					System.Text.StringBuilder buffer = new System.Text.StringBuilder();
					buffer.Append(NameToString(item.Name));
					buffer.Append("(");
					for(int i = 0; i < item.Arity; i++)
					{
						if (i > 0)
						{
							buffer.Append(",");
						}
						buffer.Append(CellToString(item[i]));
					}
					buffer.Append(")");
					return buffer.ToString();
				}
			}
		}
		
		public override string ToString()
		{
			System.Text.StringBuilder buffer = new System.Text.StringBuilder();
			for(int i = 0; i < Length; i++)
			{
				buffer.Append(CellToString(this[i]));
				buffer.AppendLine(".");
			}
			return buffer.ToString();
		}
	}
}
