using System;
using System.Collections;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;
using System.Drawing;

namespace Quest.Graph2D
{
	[Serializable]
	public class GraphLayout
	{
		[Serializable]
		internal struct NodeLayout: ISerializable
		{
			public const int Flag_Pin = 01;

			public string ID;
			public Point Position;
			public int Flags;

			#region Implementation of ISerializable
			private const string SID_ID = "ID";
			private const string SID_Position = "Position";
			private const string SID_Flags = "Flags";

			// field addition tolerant
			public NodeLayout(SerializationInfo info, StreamingContext context)
			{
				ID = "";
				Position = new Point();
				Flags = 0;
				foreach(SerializationEntry entry in info)
				{
					switch(entry.Name)
					{
						case SID_ID:
							ID = (string)entry.Value;
							break;

						case SID_Position:
							Position = (Point)entry.Value;
							break;

						case SID_Flags:
							Flags = (int)entry.Value;
							break;
					}
				}
			}

			public void GetObjectData(SerializationInfo info, StreamingContext context)
			{
				info.AddValue(SID_ID, ID);
				info.AddValue(SID_Position, Position);
				if(Flags != 0)
				{
					info.AddValue(SID_Flags, Flags);
				}
			}
			#endregion
		}

		public ArrayList Nodes = new ArrayList();

		public void Add(Node node)
		{
			NodeLayout layout = new NodeLayout();
			layout.Position = node.Position;
			layout.ID = node.ID;
			if(node.Pin)
			{
				layout.Flags |= NodeLayout.Flag_Pin;
			}
			Nodes.Add(layout);
		}
	}

	public class Graph: IGraph
	{
		#region Layout persistence
		public GraphLayout Layout
		{
			get
			{
				GraphLayout result = new GraphLayout();
				foreach(Node node in Nodes)
				{
					result.Add(node);
				}
				return result;
			}
			set
			{
				// TODO: create faster version of Nodes.BringToFront and use it in Nodes.BeginUpdate/EndUpdate mode
				Hashtable index = new Hashtable();
				foreach(Node node in Nodes)
				{
					index[node.ID] = node;
				}
				foreach(GraphLayout.NodeLayout layout in value.Nodes)
				{
					Node node = (Node)index[layout.ID];
					if(node != null)
					{
						node.Position = layout.Position;
						node.Pin = (layout.Flags & GraphLayout.NodeLayout.Flag_Pin) != 0;
						Nodes.BringToFront(node);
					}
				}
				Invalidate();
			}
		}

		internal class GraphLayoutBinder: SerializationBinder
		{
			public override System.Type BindToType(string assemblyName, string typeName)
			{
				if(typeName.EndsWith(".GraphLayout"))
				{
					return typeof(GraphLayout);
				}
				else if(typeName.EndsWith(".GraphLayout+NodeLayout"))
				{
					return typeof(GraphLayout.NodeLayout);
				}
				else
				{
					return Type.GetType(String.Format("{0}, {1}", typeName, assemblyName));
				}
			}
		}

		public void LoadLayout(string fileName)
		{
			try
			{
				Stream fs = new FileStream(fileName, FileMode.Open);
				try
				{
					BinaryFormatter formatter = new BinaryFormatter();
					formatter.Binder = new GraphLayoutBinder();
					Layout = (GraphLayout)formatter.Deserialize(fs);
				}
				finally
				{
					fs.Close();
				}

			}
			catch(Exception)
			{
			}
		}

		public void SaveLayout(string fileName)
		{
			Stream fs = new FileStream(fileName, FileMode.Create);
			try
			{
				BinaryFormatter formatter = new BinaryFormatter();
				formatter.Serialize(fs, Layout);
			}
			finally
			{
				fs.Close();
			}
		}
		#endregion

		public NodeCollection Nodes;
		public EdgeCollection Edges;

		public Graph()
		{
			Nodes = new NodeCollection(this);
			Edges = new EdgeCollection(this);
		}

		protected Space pSpace = null;
		protected Space space
		{
			get
			{
				if(pSpace == null)
				{
					pSpace = new Space(this);
				}
				return pSpace;
			}
			set
			{
				pSpace = value;
			}
		}

		public object HitTest(int anX, int aY)
		{
			// TODO: space is somehow unreliable
			return space.HitTest(anX, aY);
			// return Nodes.HitTest(anX, aY);
		}

		public void Calibrate(Graphics canvas)
		{
			Edges.Calibrate(canvas);
			Nodes.Calibrate(canvas);
			Invalidate();
		}

		public void PaintTo(Graphics canvas)
		{
			Edges.PaintTo(canvas);
			Nodes.PaintTo(canvas);
		}

		public void PaintTo(Graphics canvas, Rectangle clipRect)
		{
			foreach(Edge edge in space.SelectEdges(clipRect, true))
			{
				edge.PaintTo(canvas);
			}
			foreach(Node node in SelectNodes(clipRect, true))
			{
				node.PaintTo(canvas);
			}
		}

		public INodeList SelectNodes(Rectangle clipper, bool orderingMatters)
		{
			// TODO: space is somehow unreliable
			return space.SelectNodes(clipper, orderingMatters);
			// return Nodes;
		}

		public INodeList NeighboursOf(Node root)
		{
			// TODO:
			// return space.SelectEdges(root.Bounds /* TODO: */, false).NeighboursOf(root);
			return Edges.NeighboursOf(root);
		}

		private int updateLevel = 0;
		private bool invalidateUpdateRequest;

		public void BeginUpdate()
		{
			if(updateLevel == 0)
			{
				invalidateUpdateRequest = false;
				Edges.BeginUpdate();
			}
			updateLevel++;
		}

		public void EndUpdate()
		{
			updateLevel--;
			if(updateLevel == 0)
			{
				if (invalidateUpdateRequest)
				{
					Invalidate();
				}
				Edges.EndUpdate();
			}
		}

		public void Invalidate()
		{
			if (updateLevel == 0)
			{
				Nodes.Invalidate();
				Edges.Invalidate();
				space = null;
			}
			else
			{
				invalidateUpdateRequest = true;
			}
		}

		public Rectangle Bounds
		{
			get
			{
				return Nodes.Bounds;
			}
		}

		private EdgeLayout pDefaultEdgeLayout = null;
		public Edge New(Node nodeFrom, string key, Node nodeTo)
		{
			if(nodeFrom == null) return null;
			if(nodeTo == null) return null;
			Edge result = new Edge();
			result.From = nodeFrom;
			result.To = nodeTo;
			result.Key = key;

			if(pDefaultEdgeLayout == null)
			{
				pDefaultEdgeLayout = new EdgeLayout();
			}

			result.Layout = pDefaultEdgeLayout;
			Edges.Add(result);
			Invalidate();
			return result;
		}

		public void Remove(Node nodeFrom, string key, Node nodeTo)
		{
			Edges.Remove(nodeFrom, key, nodeTo);
			Invalidate();
		}

		public void Remove(Node nodeFrom)
		{
			foreach(Edge edge in Edges.Find(nodeFrom))
			{
				Edges.Remove(edge);
				Edges.Invalidate();
			}
			Nodes.Remove(nodeFrom);
			Invalidate();
		}

		private Relaxer relaxer = null;
		public void Relax()
		{
			if(relaxer == null)
			{
				relaxer = new Relaxer();
				relaxer.Graph = this;
			}
			relaxer.Relax();
		}

		public Point NormalizePositions()
		{
			Rectangle bounds = Bounds;
			Point result = new Point(-(bounds.Left + bounds.Width / 2), -(bounds.Top + bounds.Height / 2));
			foreach(Node node in Nodes)
			{
				node.Position.X += result.X;
				node.Position.Y += result.Y;
			}
			Invalidate();
			return result;
		}
	}
}
