using System;
using System.Collections;
using System.Diagnostics;
using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Drawing2D;
using Quest.Graph2D;

namespace Quest.Graph2D.Gui
{
	public class ViewPosition: ICloneable
	{
		/*protected internal*/public Point GraphCenter = new Point(0, 0);
		/*protected internal*/public float Zoom = 1f;

		public object Clone()
		{
			ViewPosition result = new ViewPosition();
			result.Zoom = Zoom;
			result.GraphCenter = GraphCenter;
			return result;
		}

		public override bool Equals(Object obj)
		{
			if((obj == null) || (GetType() != obj.GetType())) return false;
			ViewPosition p = (ViewPosition)obj;
			return Zoom.Equals(p.Zoom) && GraphCenter.Equals(p.GraphCenter);
		}

		public override int GetHashCode()
		{
			return Zoom.GetHashCode() ^ GraphCenter.GetHashCode();
		}
	}

	public class View: UserControl
	{
		public ViewPosition Position
		{
			get
			{
				ViewPosition result = new ViewPosition();
				result.Zoom = Zoom;
				result.GraphCenter = GraphCenter;
				return result;
			}
			set
			{
				GraphCenter = value.GraphCenter;
				Zoom = value.Zoom;
			}
		}

		private class SelectedNodesCollection: NodeCollectionBase
		{
			private View pOwner;

			public SelectedNodesCollection(View anOwner)
			{
				pOwner = anOwner;
			}

			protected override void Changed()
			{
				base.Changed();
				pOwner.Invalidate();
			}

			protected override void Link(Node value, int index)
			{
				base.Link(value, index);
				value.Selected = true;
			}

			protected override void Unlink(Node value)
			{
				base.Unlink(value);
				value.Selected = false;
			}

			public override int Add(object value)
			{
				if(value == null) return -1;

				int result = IndexOf(value);
				if(result < 0)
				{
					result = base.Add(value);
				}
				return result;
			}
		}

		private Graph pGraph = null;

		private VScrollBar verticalScrollbar;
		private HScrollBar horizontalScrollbar;

		private bool IsWindowsOS()
		{
			// TODO: how to fix Mono defficiencies ?
			string root = (string)System.Environment.SystemDirectory;
			if(root.Length >= 3)
			{
				return root.Substring(1, 2) == ":\\";
			}
			else
			{
				return false;
			}
		}

		#region Change tracking
		private bool pModified = false;
		private ViewPosition pOldPosition;
		protected void Changed()
		{
			pModified = true;
		}
		public bool Modified
		{
			get
			{
				return pModified || ((pOldPosition != null) && (!Position.Equals(pOldPosition)));
			}
			set
			{
				if (value)
				{
					pModified = true;
				}
				else
				{
					pOldPosition = Position;
					pModified = false;
				}
			}
		}
		#endregion

		public View()
		{
			// TUNE:
			if(IsWindowsOS())
			{
				SetStyle(ControlStyles.UserPaint | ControlStyles.AllPaintingInWmPaint | ControlStyles.DoubleBuffer, true);
			}
			ResizeRedraw = true;
			pSelectedNodes = new SelectedNodesCollection(this);
			InitializeComponent();
		}

		private void InitializeComponent()
		{
			verticalScrollbar = new VScrollBar();
			verticalScrollbar.Dock = DockStyle.Right;
			verticalScrollbar.TabStop = false;
			verticalScrollbar.Enabled = false;
			verticalScrollbar.Visible = true;
			verticalScrollbar.ValueChanged += new EventHandler(this.Scrollbar_ValueChanged);
			Controls.Add(verticalScrollbar);

			horizontalScrollbar = new HScrollBar();
			horizontalScrollbar.Dock = DockStyle.Bottom;
			horizontalScrollbar.TabStop = false;
			horizontalScrollbar.Enabled = false;
			horizontalScrollbar.Visible = true;
			horizontalScrollbar.ValueChanged += new EventHandler(this.Scrollbar_ValueChanged);
			Controls.Add(horizontalScrollbar);
		}

		#region Public properties used at run time
		public Graph Graph
		{
			get
			{
				return pGraph;
			}
			set
			{
				pGraph = value;
				if(pGraph != null)
				{
					Graphics canvas = CreateGraphics();
					try
					{
						pGraph.Calibrate(canvas);
					}
					finally
					{
						canvas.Dispose();
					}

					GraphCenter = Center(pGraph.Bounds);
				}
				else
				{
					UpdateScrollbars();
				}
				Modified = false;
				Invalidate();
			}
		}

		private SelectedNodesCollection pSelectedNodes;
		public INodeList SelectedNodes
		{
			get
			{
				return pSelectedNodes;
			}
			set
			{
				pSelectedNodes.Clear();
				foreach(Node node in value)
				{
					pSelectedNodes.Add(node);
				}
			}
		}

		public Node SelectedNode
		{
			get
			{
				return (pSelectedNodes.Count == 1) ? pSelectedNodes[0] : null;
			}
			set
			{
				pSelectedNodes.Clear();
				pSelectedNodes.Add(value);
			}
		}
		#endregion

		public void FocusNode(Node node, bool select)
		{
			GraphCenter = node.Position;
			if(select)
			{
				SelectedNode = node;
				if (pGraph.Nodes.BringToFront(node, true))
				{
					Invalidate();
				}
			}
			else
			{
				Invalidate(); // TODO: this should be covered elsewhere (GraphCenter.set)
			}
		}

		public void GraphChanged()
		{
			pGraph.Invalidate();
			UpdateScrollbars();
			Invalidate();
			Changed();
		}

		public void GraphChanged(Node newNode)
		{
			GraphChanged(new Node[]{newNode});
		}

		public void GraphChanged(Edge newEdge)
		{
			GraphChanged(new Edge[]{newEdge});
		}

		public void GraphChanged(Edge[] newEdges)
		{
			Graphics canvas = CreateGraphics();
			try
			{
				foreach (Edge edge in newEdges)
				{
					if (edge != null)
					{
						edge.Calibrate(canvas);
					}
				}
			}
			finally
			{
				canvas.Dispose();
			}
			GraphChanged();
		}

		public void GraphChanged(IEdgeList newEdges)
		{
			Graphics canvas = CreateGraphics();
			try
			{
				foreach (Edge edge in newEdges)
				{
					if (edge != null)
					{
						edge.Calibrate(canvas);
					}
				}
			}
			finally
			{
				canvas.Dispose();
			}
			GraphChanged();
		}

		public void GraphChanged(Node[] newNodes)
		{
			Graphics canvas = CreateGraphics();
			try
			{
				foreach(Node node in newNodes)
				{
					node.Calibrate(canvas);
				}
			}
			finally
			{
				canvas.Dispose();
			}
			GraphChanged();
		}

		public void GraphChanged(INodeList newNodes)
		{
			Graphics canvas = CreateGraphics();
			try
			{
				foreach(Node node in newNodes)
				{
					node.Calibrate(canvas);
				}
			}
			finally
			{
				canvas.Dispose();
			}
			GraphChanged();
		}

		protected Point Center(Rectangle rect)
		{
			return new Point
				(
				 rect.Left + (rect.Width / 2),
				 rect.Top + (rect.Height / 2)
				);
		}

		protected object HitTest(int anX, int aY)
		{
			return (pGraph == null) ? null : pGraph.HitTest(anX, aY);
		}

		private Point pGraphCenter = new Point(0, 0);
		protected Point GraphCenter
		{
			get
			{
				return pGraphCenter;
			}
			set
			{
				pGraphCenter = value;
				UpdateScrollbars();
				// TODO: reflect changes
			}
		}

		public void NormalizePositions()
		{
			Point shift = pGraph.NormalizePositions();
			pGraphCenter.X += shift.X;
			pGraphCenter.Y += shift.Y;
			UpdateScrollbars();
			Invalidate();
		}

		protected Point PictureCenter
		{
			get
			{
				return GraphCoordinateToPictureCoordinate(GraphCenter);
			}
		}

		public Point PointToGraph(Point p)
		{
			return ViewportCoordinateToGraphCoordinate(p);
		}

		// Viewport is a rectangular area within control, displaying
		// picture. Picture's center is located at Viewport's center
		protected Rectangle ViewportBounds
		{
			get
			{
				Rectangle client = ClientRectangle;
				return new Rectangle
					(
					 client.Left,
					 client.Top,
					 client.Width - verticalScrollbar.Width,
					 client.Height - horizontalScrollbar.Height
					);
			}
		}

		protected Point ViewportCenter
		{
			get
			{
				return Center(ViewportBounds);
			}
			set
			{
				Point oldCenter = ViewportCenter;
				Size centerDistance = new Size(value.X - oldCenter.X, value.Y - oldCenter.Y);
				centerDistance = ViewportDistanceToGraphDistance(centerDistance);
				GraphCenter = new Point(GraphCenter.X + centerDistance.Width, GraphCenter.Y + centerDistance.Height);
				// TODO:
				UpdateScrollbars();
				Invalidate();
			}
		}

		private Matrix pTransform = new Matrix();
		private Matrix pInvertedTransform = new Matrix();
		protected Matrix Transform
		{
			get
			{
				return pTransform;
			}
			set
			{
				pTransform = (Matrix)value.Clone();
				pInvertedTransform = (Matrix)value.Clone();
				pInvertedTransform.Invert();
				if(pGraph != null)
				{
					UpdateScrollbars();
					Invalidate();
					// TODO: rescale ?
				}
			}
		}

		private float pZoom = 1;
		protected float Zoom
		{
			get
			{
				return pZoom;
			}
			set
			{
				pZoom = value;
				Matrix m = new Matrix();
				m.Scale(pZoom, pZoom);
				Transform = m;
			}
		}

		#region Coordinate translations
		protected Point GraphCoordinateToPictureCoordinate(Point p)
		{
			Point[] points = new Point[]{p};
			pTransform.TransformPoints(points);
			return points[0];
		}

		protected Size GraphDistanceToPictureDistance(Size d)
		{
			Point p = new Point(d.Width, d.Height);
			p = GraphCoordinateToPictureCoordinate(p);
			return new Size(p.X, p.Y);
		}

		protected Rectangle GraphRectToPictureRect(Rectangle r)
		{
			return new Rectangle
				(
				 GraphCoordinateToPictureCoordinate(new Point(r.Left, r.Top)),
				 GraphDistanceToPictureDistance(new Size(r.Width, r.Height))
				);
		}

		protected Rectangle PictureRectToGraphRect(Rectangle r)
		{
			return new Rectangle
				(
				 PictureCoordinateToGraphCoordinate(new Point(r.Left, r.Top)),
				 PictureDistanceToGraphDistance(new Size(r.Width, r.Height))
				);
		}

		protected Rectangle PictureRectToViewportRect(Rectangle r)
		{
			return new Rectangle
				(
				 PictureCoordinateToViewportCoordinate(new Point(r.Left, r.Top)),
				 PictureDistanceToViewportDistance(new Size(r.Width, r.Height))
				);
		}

		protected Rectangle ViewportRectToPictureRect(Rectangle r)
		{
			return new Rectangle
				(
				 ViewportCoordinateToPictureCoordinate(new Point(r.Left, r.Top)),
				 ViewportDistanceToPictureDistance(new Size(r.Width, r.Height))
				);
		}

		protected Rectangle GraphRectToViewportRect(Rectangle r)
		{
			return PictureRectToViewportRect(GraphRectToPictureRect(r));
		}

		protected Rectangle ViewportRectToGraphRect(Rectangle r)
		{
			return PictureRectToGraphRect(ViewportRectToPictureRect(r));
		}

		protected Point PictureCoordinateToGraphCoordinate(Point p)
		{
			Point[] points = new Point[]{p};
			pInvertedTransform.TransformPoints(points);
			return points[0];
		}

		protected Size PictureDistanceToGraphDistance(Size d)
		{
			Point p = new Point(d.Width, d.Height);
			p = PictureCoordinateToGraphCoordinate(p);
			return new Size(p.X, p.Y);
		}

		protected Point ViewportCoordinateToPictureCoordinate(Point p)
		{
			Point pictureCenter = PictureCenter;
			Point viewportCenter = ViewportCenter;
			return new Point
				(
				 p.X - viewportCenter.X + pictureCenter.X,
				 p.Y - viewportCenter.Y + pictureCenter.Y
				);
		}

		protected Size ViewportDistanceToPictureDistance(Size d)
		{
			return d;
		}

		protected Size PictureDistanceToViewportDistance(Size d)
		{
			return d;
		}

		protected Point PictureCoordinateToViewportCoordinate(Point p)
		{
			Point pictureCenter = PictureCenter;
			Point viewportCenter = ViewportCenter;
			return new Point
				(
				 p.X - pictureCenter.X + viewportCenter.X,
				 p.Y - pictureCenter.Y + viewportCenter.Y
				);
		}

		protected Point ViewportCoordinateToGraphCoordinate(Point p)
		{
			return PictureCoordinateToGraphCoordinate(ViewportCoordinateToPictureCoordinate(p));
		}

		protected Size ViewportDistanceToGraphDistance(Size d)
		{
			return PictureDistanceToGraphDistance(ViewportDistanceToPictureDistance(d));
		}

		protected Point MouseCoordinateToGraphCoordinate(Point p)
		{
			return ViewportCoordinateToGraphCoordinate(p);
		}

		protected Size MouseDistanceToGraphDistance(Size d)
		{
			return ViewportDistanceToGraphDistance(d);
		}
		#endregion

		protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
		{
			base.OnPaint(e);
			if(pGraph != null)
			{
				Matrix m = new Matrix();
				Point offset = PictureCoordinateToViewportCoordinate(new Point(0,0));
				m.Translate(offset.X, offset.Y);
				m.Multiply(pTransform);
				e.Graphics.Transform = m;

				Pen p = new Pen(Color.Silver);
				e.Graphics.DrawRectangle(p, pGraph.Bounds);

				pGraph.PaintTo(e.Graphics, ViewportRectToGraphRect(ViewportBounds));
			}
		}

		#region Mouse Moves
		private enum MouseMoveMode
		{
			None,
			NodeMoving,
			GraphMoving
		}

		private MouseMoveMode pMoveMode = MouseMoveMode.None;

		private Node pPickedNode = null;
		private Point pPickedShift;
		private void MovePicker(int aX, int anY)
		{
			Point delta = new Point(aX + pPickedShift.X - pPickedNode.Position.X, anY + pPickedShift.Y - pPickedNode.Position.Y);
			if((delta.X != 0) || (delta.Y != 0))
			{
				foreach(Node node in SelectedNodes)
				{
					node.Position = new Point(node.Position.X + delta.X, node.Position.Y + delta.Y);
				}
				// TODO: scroll picked node to view
				pGraph.Invalidate();
				UpdateScrollbars();
				Invalidate();
				Changed();
			}
		}

		private void MakeVisible(int aX, int anY)
		{
			// TUNE:
			const int hitX = 5;
			const int hitY = 5;

			Rectangle bounds = ViewportBounds;
			int dx = 0;
			int dy = 0;
			if(aX - hitX < bounds.Left)
			{
				dx = (aX - hitX) - bounds.Left;
			}
			else if(aX + hitX > bounds.Right)
			{
				dx = (aX + hitX) - bounds.Right;
			}
			if(anY - hitY < bounds.Top)
			{
				dy = (anY - hitY) - bounds.Top;
			}
			else if(anY + hitY > bounds.Bottom)
			{
				dy = (anY + hitY) - bounds.Bottom;
			}
			if((dx != 0) || (dy != 0))
			{
				Point oldCenter = ViewportCenter;
				ViewportCenter = new Point(oldCenter.X + dx, oldCenter.Y + dy);
			}
		}

		private bool pCursorSet;
		private Point pMouseZero;
		private Point pGraphZero;
		private void MoveGraph(Point newMouse)
		{
			Size dist = MouseDistanceToGraphDistance(new Size(newMouse.X - pMouseZero.X, newMouse.Y - pMouseZero.Y));
			Point newGraphCenter = new Point(pGraphZero.X - dist.Width, pGraphZero.Y - dist.Height);
			if(!newGraphCenter.Equals(GraphCenter))
			{
				GraphCenter = newGraphCenter;
				Invalidate();
			}
		}

		protected override void OnDoubleClick(System.EventArgs e)
		{
			Point p = PointToClient(Control.MousePosition);
			if(!PopupContextMenu(p.X, p.Y, true))
			{
				base.OnDoubleClick(e);
			}
		}

		protected static bool KeyPressed(Keys key)
		{
			return (Control.ModifierKeys & key) == key;
		}

		protected override void OnMouseDown(System.Windows.Forms.MouseEventArgs e)
		{
			base.OnMouseDown(e);
			if(pGraph == null)
			{
			}
			else if(e.Button == MouseButtons.Left)
			{
				Point p = MouseCoordinateToGraphCoordinate(new Point(e.X, e.Y));
				pPickedNode = (Node)HitTest(p.X, p.Y);
				if(pPickedNode != null)
				{
					if(SelectedNodes.Contains(pPickedNode) && KeyPressed(Keys.Shift))
					{
						SelectedNodes.Remove(pPickedNode);
					}
					else
					{
						pMoveMode = MouseMoveMode.NodeMoving;

						pPickedShift = new Point(pPickedNode.Position.X - p.X, pPickedNode.Position.Y - p.Y);
						if(pGraph.Nodes.BringToFront(pPickedNode, true))
						{
							Invalidate();
						}

						if(!SelectedNodes.Contains(pPickedNode))
						{
							if(KeyPressed(Keys.Shift))
							{
								SelectedNodes.Add(pPickedNode);
							}
							else
							{
								SelectedNode = pPickedNode;
							}
						}
					}
				}
				else
				{
					SelectedNodes.Clear();

					pMouseZero = new Point(e.X, e.Y);
					pGraphZero = GraphCenter;
					pMoveMode = MouseMoveMode.GraphMoving;

					pCursorSet = false;
				}
			}
			else if(e.Button == MouseButtons.Middle)
			{
				pMouseZero = new Point(e.X, e.Y);
				pGraphZero = GraphCenter;
				pMoveMode = MouseMoveMode.GraphMoving;

				Cursor = Cursors.SizeAll;
				pCursorSet = true;
			}
			else if(e.Button == MouseButtons.Right)
			{
				Point p = MouseCoordinateToGraphCoordinate(new Point(e.X, e.Y));
				Node pickedNode = (Node)HitTest(p.X, p.Y);
				if(!SelectedNodes.Contains(pickedNode))
				{
					SelectedNode = pickedNode;
				}
				PopupContextMenu(e.X, e.Y, false);
			}
		}

		protected override void OnMouseMove(System.Windows.Forms.MouseEventArgs e)
		{
			base.OnMouseMove(e);
			switch(pMoveMode)
			{
				case MouseMoveMode.NodeMoving:
					{
						MakeVisible(e.X, e.Y);
						Point p = MouseCoordinateToGraphCoordinate(new Point(e.X, e.Y));
						MovePicker(p.X, p.Y);
						MakeVisible(e.X, e.Y);
						break;
					}

				case MouseMoveMode.GraphMoving:
					{
						if(!pCursorSet)
						{
							Cursor = Cursors.SizeAll;
							pCursorSet = true;
						}
						MoveGraph(new Point(e.X, e.Y));
						break;
					}
			}
		}

		protected override void OnMouseUp(System.Windows.Forms.MouseEventArgs e)
		{
			base.OnMouseUp(e);
			switch(pMoveMode)
			{
				case MouseMoveMode.NodeMoving:
					{
						/* TODO: this should have already been handled in OnMouseMove
						MakeVisible(e.X, e.Y);
						Point p = MouseCoordinateToGraphCoordinate(new Point(e.X, e.Y));
						MovePicker(p.X, p.Y);
						MakeVisible(e.X, e.Y);
						*/
						pPickedNode = null;
						break;
					}

				case MouseMoveMode.GraphMoving:
					{
						/* TODO: this should have already been handled in OnMouseMove
						MoveGraph(new Point(e.X, e.Y));
						*/
						Cursor = null;
						break;
					}
			}
			pMoveMode = MouseMoveMode.None;
		}
		#endregion

		private ContextMenu pContextMenu = null;
		public override ContextMenu ContextMenu
		{
			get
			{
				return pContextMenu;
			}
			set
			{
				pContextMenu = value;
			}
		}

		private Node pSourceNode = null;
		public Node SourceNode
		{
			get
			{
				return pSourceNode;
			}
		}

		private bool PopupContextMenu(int anX, int aY, bool onlyDefault)
		{
			if(pContextMenu == null) return false;
			Point controlCoord = new Point(anX, aY);
			if(ViewportBounds.Contains(controlCoord))
			{
				Point graphCoord = ViewportCoordinateToGraphCoordinate(controlCoord);
				pSourceNode = (Node)HitTest(graphCoord.X, graphCoord.Y);
			}
			else
			{
				pSourceNode = null;
			}
			if(onlyDefault)
			{
				foreach(MenuItem item in pContextMenu.MenuItems)
				{
					if(item.DefaultItem)
					{
						// TODO: OnPopup's should be notified
						item.PerformClick();
						return true;
					}
				}
				return false;
			}
			else
			{
				pContextMenu.Show(this, controlCoord);
			}
			return true;
		}

		// TUNE:
		private const float ZoomFactor = 1.2f;
		private void ZoomIn(bool marginal)
		{
			if(marginal)
			{
				Zoom = 1;
			}
			else
			{
				Zoom = Zoom * ZoomFactor;
			}
		}

		private void ZoomOut(bool marginal, Rectangle rect)
		{
			if(marginal)
			{

				while(!ViewportBounds.Contains(GraphRectToViewportRect(rect)))
				{
					ZoomOut(false);
				}
			}
			else
			{
				Zoom = Zoom * (1 / ZoomFactor);
			}
		}

		private void ZoomOut(bool marginal)
		{
			ZoomOut(marginal, pGraph.Bounds);
		}

		protected override void OnKeyPress(System.Windows.Forms.KeyPressEventArgs e)
		{
			base.OnKeyPress(e);
			switch(e.KeyChar)
			{
				case '+':
					ZoomIn(KeyPressed(Keys.Shift));
					break;

				case '-':
					ZoomOut(KeyPressed(Keys.Shift));
					break;
			}
		}

		protected override void OnMouseWheel(System.Windows.Forms.MouseEventArgs e)
		{
			// NOTE: suppress inherited behaviour
			// base.OnMouseWheel(e);
			Point viewportCoordinate = new Point(e.X, e.Y);
			if(ViewportBounds.Contains(viewportCoordinate) && (pMoveMode == MouseMoveMode.None))
			{
				Point oldGraphCoordinate = ViewportCoordinateToGraphCoordinate(viewportCoordinate);

				if(e.Delta > 0)
				{
					ZoomIn(KeyPressed(Keys.Shift));
				}
				else
				{
					if(pSelectedNodes.Count > 0)
					{
						ZoomOut(KeyPressed(Keys.Shift), pSelectedNodes.Bounds);
					}
					else
					{
						ZoomOut(KeyPressed(Keys.Shift));
					}
				}

				// select new focus point so that mouse cursor points to the same graph point
				Point newGraphCoordinate = ViewportCoordinateToGraphCoordinate(viewportCoordinate);
				GraphCenter = new Point
					(
					 GraphCenter.X - newGraphCoordinate.X + oldGraphCoordinate.X,
					 GraphCenter.Y - newGraphCoordinate.Y + oldGraphCoordinate.Y
					);
				Invalidate();
			}
		}

		protected override void OnResize(System.EventArgs e)
		{
			base.OnResize(e);
			UpdateScrollbars();
		}

		private void UpdateScrollbar(ScrollBar bar, int sign, bool largeStep)
		{
			if(!bar.Enabled) return;
			int newValue;
			if(largeStep)
			{
				newValue = bar.Value + (bar.LargeChange * sign);
			}
			else
			{
				newValue = bar.Value + (bar.SmallChange * sign);
			}
			if(newValue < bar.Minimum) newValue = bar.Minimum;
			if(newValue > bar.Maximum) newValue = bar.Maximum;
			if(newValue != bar.Value) bar.Value = newValue;
		}

		private bool ProcessKey(Keys key, bool down)
		{
			switch(key)
			{
				case Keys.Down:
					if(down) UpdateScrollbar(verticalScrollbar, +1, false);
					return true;

				case Keys.Up:
					if(down) UpdateScrollbar(verticalScrollbar, -1, false);
					return true;

				case Keys.Left:
					if(down) UpdateScrollbar(horizontalScrollbar, -1, false);
					return true;

				case Keys.Right:
					if(down) UpdateScrollbar(horizontalScrollbar, +1, false);
					return true;

				case Keys.Down | Keys.Control:
					if(down) UpdateScrollbar(verticalScrollbar, +1, true);
					return true;

				case Keys.Up | Keys.Control:
					if(down) UpdateScrollbar(verticalScrollbar, -1, true);
					return true;

				case Keys.Left | Keys.Control:
					if(down) UpdateScrollbar(horizontalScrollbar, -1, true);
					return true;

				case Keys.Right | Keys.Control:
					if(down) UpdateScrollbar(horizontalScrollbar, +1, true);
					return true;

				case Keys.PageDown:
				case Keys.PageUp:
					return true;

				default:
					return false;
			}
		}

		protected override void OnKeyDown(System.Windows.Forms.KeyEventArgs e)
		{
			if(ProcessKey(e.KeyCode, true)) e.Handled = true;
			if(!e.Handled) base.OnKeyDown(e);
		}

		protected override void OnKeyUp(System.Windows.Forms.KeyEventArgs e)
		{
			if(ProcessKey(e.KeyCode, false)) e.Handled = true;
			if(!e.Handled) base.OnKeyDown(e);
		}

		protected override bool ProcessDialogKey(System.Windows.Forms.Keys keyData)
		{
			if(ProcessKey(keyData, true))
				return true;
			return base.ProcessDialogKey(keyData);
		}

		private bool pEnableScroll = true;
		private void Scrollbar_ValueChanged(Object sender, EventArgs e)
		{
			if(!pEnableScroll) return;
			Point newCenter = new Point
				(
				 horizontalScrollbar.Enabled ? horizontalScrollbar.Value : GraphCenter.X,
				 verticalScrollbar.Enabled ? verticalScrollbar.Value : GraphCenter.Y
				);
			if(!GraphCenter.Equals(newCenter))
			{
				pEnableScroll = false;
				try
				{
					GraphCenter = newCenter;
					Invalidate();
				}
				finally
				{
					pEnableScroll = true;
				}
			}
		}

		private void UpdateScrollbar(ScrollBar bar, int minValue, int maxValue, int value)
		{
			if(value < minValue) minValue = value;
			if(value > maxValue) maxValue = value;
			if(minValue >= maxValue)
			{
				if(bar.Enabled) bar.Enabled = false;
			}
			else
			{
				if((minValue != bar.Minimum) || (maxValue != bar.Maximum) || (value != bar.Value))
				{
					// Debug.WriteLine(String.Format("{0} {1} <= {2} <= {3}", bar.GetType().Name, minValue, value, maxValue));
					if((minValue != bar.Minimum) || (maxValue != bar.Maximum))
					{
						bar.Minimum = Int32.MinValue;
						bar.Maximum = Int32.MaxValue;
						bar.Value = value;
						bar.Minimum = minValue;
						bar.Maximum = maxValue;
					}
					else
					{
						bar.Value = value;
					}
				}
				if(!bar.Enabled) bar.Enabled = true;
			}
		}

		protected void UpdateScrollbars()
		{
			pEnableScroll = false;
			try
			{
				if(pGraph == null)
				{
					verticalScrollbar.Enabled = false;
					horizontalScrollbar.Enabled = false;
				}
				else
				{
					/*
					// TODO: calculate smart scrollbars
					Rectangle viewport = ViewportBounds;
					Size graphWindow = ViewportDistanceToGraphDistance(viewport.Size);
					Rectangle viewportIntoGraph	= new Rectangle
						(
						 ViewportCoordinateToGraphCoordinate(viewport.Location),
						 graphWindow
						);

					Rectangle optimalGraphBounds = Rectangle.Union(viewportIntoGraph, Rectangle.Union(pGraph.Bounds, new Rectangle(GraphCenter, new Size(1, 1))));
					int graphWindowDX = (graphWindow.Width / 2);
					int graphWindowDY = (graphWindow.Height / 2);

					optimalGraphBounds = new Rectangle
						(
						 optimalGraphBounds.Left + graphWindowDX,
						 optimalGraphBounds.Top + graphWindowDY,
						 optimalGraphBounds.Width - (2 * graphWindowDX),
						 optimalGraphBounds.Height - (2 * graphWindowDY)
						);
					 */
					Rectangle optimalGraphBounds = pGraph.Bounds;

					UpdateScrollbar
						(
						 horizontalScrollbar,
						 optimalGraphBounds.Left,
						 optimalGraphBounds.Right,
						 GraphCenter.X
						);

					UpdateScrollbar
						(
						 verticalScrollbar,
						 optimalGraphBounds.Top,
						 optimalGraphBounds.Bottom,
						 GraphCenter.Y
						);
				}
			}
			finally
			{
				pEnableScroll = true;
			}
		}
	}
}
