using System;
using System.Collections;
using System.Drawing;

namespace Quest.Graph2D
{
	public interface INodeList: IList, ICloneable
	{
		new Node this[int index] { get; }
		INodeList Select(string filterExpression);
	}

	public class NodeLayout: ICloneable
	{
		public Font Font;
		public Brush TextStyle;
		public Brush BackgroundStyle;
		public Pen BorderStyle;

		public object Clone()
		{
			NodeLayout result = new NodeLayout();
			result.Font = Font;
			result.TextStyle = TextStyle;
			result.BackgroundStyle = BackgroundStyle;
			result.BorderStyle = BorderStyle;
			return result;
		}

		private void Setup(Color backgroundColor, Color borderColor, Font aFont, Color textColor)
		{
			if(aFont != null)
			{
				Font = (Font)aFont.Clone();
			}
			if(backgroundColor != Color.Empty)
			{
				BackgroundStyle = new SolidBrush(backgroundColor);
			}
			if(textColor != Color.Empty)
			{
				TextStyle = new SolidBrush(textColor);
			}
			if(borderColor != Color.Empty)
			{
				BorderStyle = new Pen(borderColor);
			}
		}

		public NodeLayout(Color backgroundColor, Color borderColor, Font aFont, Color textColor)
		{
			Setup(backgroundColor, borderColor, aFont, textColor);
		}

		public NodeLayout(Color backgroundColor, Color borderColor, Color textColor)
		{
			Setup(backgroundColor, borderColor, System.Windows.Forms.Control.DefaultFont, textColor);
		}

		public NodeLayout()
		{
			Setup(Color.White, Color.Black, System.Windows.Forms.Control.DefaultFont, Color.Black);
		}
	}

	public class Node: IGraphElement
	{
		// TUNE:
		private const int HorizontalBorder = 2;
		private const int VerticalBorder = 1;

		public Point Position = new Point(0, 0);
		public Size Size = new Size(0, 0);

		private string pLabel = "";
		protected string[] pLabelLines = new string[0];
		protected bool[] pLabelGroups = new bool[0];
		protected int[] LabelWidth;
		protected int LabelHeight;
		public string Label
		{
			get
			{
				return pLabel;
			}
			set
			{
				pLabel = (value == null) ? "" : value;
				if(pLabel == "")
				{
					pLabelLines = new string[0];
					pLabelGroups = new bool[0];
				}
				else
				{
					string[] result = pLabel.Split(new char[]{'|'});
					pLabelGroups = new bool[result.Length];
					for(int i = 0; i < result.Length; i++)
					{
						string line = result[i].Trim();
						if((line.Length > 0) && (line[0] == '+'))
						{
							result[i] = line.Remove(0, 1);
							pLabelGroups[i] = true;
						}
						else
						{
							result[i] = line;
							pLabelGroups[i] = false;
						}
					}
					pLabelLines = result;
				}
			}
		}

		public string[] LabelLines
		{
			get
			{
				return pLabelLines;
			}
		}

		public NodeLayout Layout;

		public virtual string ReferenceID() 
		{
			return "";
		}

		public virtual string ID 
		{ 
			get
			{
				return "";
			}
		}

		public Node()
		{
		}

		public virtual void Calibrate(Graphics canvas)
		{
			SizeF textSize;
			if(Layout == null)
			{
				textSize = new SizeF();
			}
			else
			{
				textSize = canvas.MeasureString("M", Layout.Font);
				LabelHeight = (int)textSize.Height;
				if(Label != "")
				{
					LabelWidth = new int[pLabelLines.Length];
					float width = textSize.Width;
					float height = 0;
					for(int i = 0; i < pLabelLines.Length; i++)
					{
						// TODO: MeasureString differs from DrawString output e.g.
						// "pracovn pleitost v nov vybudovanm zvod v Ostrav  Hrabov"
						textSize = canvas.MeasureString(pLabelLines[i], Layout.Font);
						LabelWidth[i] = (int)textSize.Width;
						height += LabelHeight;
						if(width < textSize.Width)
						{
							width = textSize.Width;
						}
					}
					Size = new Size
						(
						 (int)(width + 2 * HorizontalBorder),
						 (int)(height + 2 * VerticalBorder)
						);

				}
				else
				{
					// TUNE: size of empty box
					float shrink = 3f / 4f;
					float radius = Math.Min((textSize.Width + 2 * HorizontalBorder) * shrink, (textSize.Height + 2 * VerticalBorder) * shrink);
					Size = new Size((int)radius, (int)radius);
				}
			}
		}

		public Rectangle Bounds
		{
			get
			{
				return new Rectangle
					(
					 Position.X - (Size.Width / 2),
					 Position.Y - (Size.Height / 2),
					 Size.Width,
					 Size.Height
					);
			}
		}

		public object HitTest(int aX, int anY)
		{
			return Bounds.Contains(aX, anY) ? this : null;
		}

		public bool Selected = false;
		public bool Pin = false;

		internal bool Mark = false;

		public static Point BoundaryHit(Rectangle bounds, int directionX, int directionY)
		{
			Point Position = new Point(bounds.Left + (bounds.Width / 2), bounds.Top + (bounds.Height / 2));

			if ((directionX == 0) && (directionY == 0))
			{
				throw new ArgumentOutOfRangeException();
			}

			Point fakeT = new Point(Position.X + directionX, Position.Y + directionY);
			while (bounds.Contains(fakeT))
			{
				fakeT.X += directionX;
				fakeT.Y += directionY;
			}

			return BoundaryHit(bounds, fakeT);
		}

		public Point BoundaryHit(int directionX, int directionY)
		{
			return BoundaryHit(Bounds, directionX, directionY);
		}

		public static Point BoundaryHit(Rectangle bounds, Point lookAt)
		{
			Point Position = new Point(bounds.Left + (bounds.Width / 2), bounds.Top + (bounds.Height / 2));
			// TODO: ? if(bounds.Contains(lookAt)) return null;

			Point F = Position;
			Point T = lookAt;
			double dx = T.X - F.X;
			double dy = T.Y - F.Y;

			if(dx != 0)
			{
				// test hitting left boundary
				{
					double t = (bounds.Left - F.X) / dx;
					if((t >= 0) && (t <= 1))
					{
						int y = (int)(F.Y + t * dy);
						if((y >= bounds.Top) && (y <= bounds.Bottom))
						{
							return new Point(bounds.Left, y);
						}
					}
				}

				// test hitting right boundary
				{
					double t = (bounds.Right - F.X) / dx;
					if((t >= 0) && (t <= 1))
					{
						int y = (int)(F.Y + t * dy);
						if((y >= bounds.Top) && (y <= bounds.Bottom))
						{
							return new Point(bounds.Right, y);
						}
					}
				}
			}

			if(dy != 0)
			{
				// test hitting top boundary
				{
					double t = (bounds.Top - F.Y) / dy;
					if((t >= 0) && (t <= 1))
					{
						int x = (int)(F.X + t * dx);
						if((x >= bounds.Left) && (x <= bounds.Right))
						{
							return new Point(x, bounds.Top);
						}
					}
				}

				// test hitting bottom boundary
				{
					double t = (bounds.Bottom - F.Y) / dy;
					if((t >= 0) && (t <= 1))
					{
						int x = (int)(F.X + t * dx);
						if((x >= bounds.Left) && (x <= bounds.Right))
						{
							return new Point(x, bounds.Bottom);
						}
					}
				}
			}

			return F; // TODO: ?
		}

		public Point BoundaryHit(Point lookAt)
		{
			return BoundaryHit(Bounds, lookAt);
		}

		public virtual void PaintTo(Graphics canvas)
		{
			if(Layout != null)
			{
				Rectangle bounds = Bounds;
				NodeLayout layout;
				if(Selected)
				{
					// TODO: TUNE: selected layout
					layout = new NodeLayout(Color.Blue, Color.Black, Color.White);
				}
				else
				{
					layout = Layout;
				}

				if(layout.BackgroundStyle != null)
				{
					canvas.FillRectangle(layout.BackgroundStyle, bounds);
				}
				if((Label != "") && (layout.Font != null) && (layout.TextStyle != null) && (GraphElement.PixelSize(canvas, LabelHeight) >= 5 /* TUNE: */))
				{
					int y = bounds.Top + VerticalBorder;
					for(int i = 0; i < pLabelLines.Length; i++)
					{
						canvas.DrawString(pLabelLines[i], layout.Font, layout.TextStyle, bounds.Left + ((bounds.Width - LabelWidth[i]) / 2), y);
						if(pLabelGroups[i])
						{
							Pen separator = (Pen)Layout.BorderStyle.Clone();
							separator.Color = Color.Gray; // TUNE:
							canvas.DrawLine(separator, bounds.Left, y, bounds.Right, y);
						}
						y += LabelHeight;
					}
				}
				{
					if(layout.BorderStyle != null)
					{
						canvas.DrawRectangle(layout.BorderStyle, bounds.Left, bounds.Top, bounds.Width, bounds.Height);
					}

					if(Pin)
					{
						// TODO: TUNE:
						int radius = 2;
						Pen p = new Pen(Selected ? Color.White : Color.Black);
						canvas.DrawEllipse(p, bounds.Right - 1 - 2 * radius - 1, bounds.Top + 1 + 1, 2 * radius, 2 * radius);
					}
				}
			}
		}

		public int Index;

		public PointF Momentum = new PointF(0, 0);

		public static bool Contains(string text, string filterExpression)
		{
			return (text == null) ? false : text.ToUpper().IndexOf(filterExpression.ToUpper()) >= 0;
		}

		public virtual bool Contains(string filterExpression)
		{
			return Contains(Label, filterExpression);
		}
	}

	internal class NodeCollectionLight: ArrayList, INodeList, IComparer
	{
		public new Node this[int index]
		{
			get
			{
				return (Node)base[index];
			}
		}

		public INodeList Select(string filterExpression)
		{
			return null; // TODO: it is suposed to be never used
		}

		public int Compare(object x, object y) 
		{
			return ((Node)x).Index - ((Node)y).Index;
		}

		public new void Sort()
		{
			base.Sort(0, Count, this);
		}
	}

	public class NodeCollectionBase: ArrayList, INodeList
	{
		public override object Clone() 
		{
			NodeCollectionBase result = new NodeCollectionBase();
			foreach(Node node in this)
			{
				result.Add(node);
			}
			return result;
		}

		protected virtual void Changed()
		{
			Invalidate();
		}

		protected virtual void Link(Node value, int index)
		{
		}

		protected virtual void Unlink(Node value)
		{
		}

		public override int Add(object value) 
		{
			int result = base.Add(value);
			Link((Node)value, result);
			Changed();
			return result;
		}

		public override void Clear() 
		{
			while(Count > 0) RemoveAt(Count - 1);
		}

		public override void Remove(object obj) 
		{
			RemoveAt(IndexOf(obj));
		}

		public override void RemoveAt(int index) 
		{
			if((index >= 0) && (index < Count))
			{
				Unlink(this[index]);
				base.RemoveAt(index);
				Changed();
			}
		}

		public new virtual Node this[int index]
		{
			get
			{
				return (Node)base[index];
			}
			set
			{
				if(value != null) Link(value, index);
				base[index] = value;
				Changed();
			}
		}

		public INodeList Select(string filterExpression)
		{
			NodeCollectionBase result = new NodeCollectionBase();
			foreach(Node node in this)
			{
				if(node.Contains(filterExpression))
				{
					result.Add(node);
				}
			}
			return result;
		}

		#region Bounds
		private bool valid = false;
		private Rectangle pBounds;
		public void Invalidate()
		{
			valid = false;
		}

		protected void Validate()
		{
			if(!valid)
			{
				valid = true;

				bool known = false;

				pBounds = Rectangle.Empty;

				foreach(Node n in this)
				{
					if(known)
					{
						pBounds = Rectangle.Union(pBounds, n.Bounds);
					}
					else
					{
						pBounds = n.Bounds;
						known = true;
					}
				}
			}
		}

		public Rectangle Bounds
		{
			get
			{
				Validate();
				return pBounds;
			}
		}
		#endregion
	}

	public class NodeCollection: NodeCollectionBase
	{
		private IGraph pGraph;

		public NodeCollection(IGraph owner)
		{
			pGraph = owner;
		}

		protected override void Link(Node value, int index) 
		{
			value.Index = index;	
		}

		public bool BringToFront(Node item)
		{
			int topIndex = Count - 1;
			if((item != null) && (item.Index != topIndex))
			{
				for(int i = item.Index + 1; i <= topIndex; i++)
				{
					this[i - 1] = this[i];
				}
				this[topIndex] = item;
				return true;
			}
			else
			{
				return false;
			}
		}

		public bool BringToFront(Node item, bool includeNeighbours)
		{
			if (!includeNeighbours)
			{
				return BringToFront(item);
			}
			else
			{
				bool result = false;
				foreach(Node neighbour in pGraph.NeighboursOf(item))
				{
					if (BringToFront(neighbour))
					{
						result = true;
					}
				}
				if (BringToFront(item))
				{
					result = true;
				}
				return result;
			}
		}

		public Node HitTest(int aX, int anY)
		{
			// we must iterate in updown order to respect visual appearance
			for(int i = Count - 1; i >= 0; i--)
			{
				Node n = this[i];
				if(n.HitTest(aX, anY) != null) return n;
			}
			return null;
		}

		public void Calibrate(Graphics canvas)
		{
			foreach(Node n in this)
			{
				n.Calibrate(canvas);
			}
			Invalidate();
		}

		public void PaintTo(Graphics canvas)
		{
			foreach(Node n in this)
			{
				n.PaintTo(canvas);
			}
		}
	}
}
