using System;
using Structure.Graph2D.IO;
using System.Xml;

namespace Structure.Graph2D.GraphML
{
	public class Writer: WriterBase
	{
		private XmlWriter _writer;
		private int _edgeId = 0;
		
		public Writer(string fileName):
			this(new System.IO.FileStream(fileName, System.IO.FileMode.Create, System.IO.FileAccess.Write))
		{
		}
		
		public Writer(System.IO.Stream stream)
		{
			XmlWriterSettings settings = new XmlWriterSettings();
			settings.CloseOutput = true;
			settings.CheckCharacters = false;
			settings.Encoding = System.Text.Encoding.UTF8;
			settings.ConformanceLevel = ConformanceLevel.Document;
			_writer = System.Xml.XmlWriter.Create(stream, settings);
			
			_writer.WriteStartElement("graphml", "http://graphml.graphdrawing.org/xmlns/graphml");
			{
				_writer.WriteAttributeString("xmlns", "y", null, "http://www.yworks.com/xml/graphml");
			}
			
			WriteElement("key", new string[]{"for", "node", "id", "d0", "yfiles.type", "nodegraphics"});
			WriteElement("key", new string[]{"for", "node", "id", "d1", "attr.name", "description", "attr.type", "string"});
			WriteElement("key", new string[]{"for", "edge", "id", "d2", "yfiles.type", "edgegraphics"});
			WriteElement("key", new string[]{"for", "edge", "id", "d3", "attr.name", "description", "attr.type", "string"});
			WriteElement("key", new string[]{"for", "edge", "id", "d4", "attr.name", "url", "attr.type", "string"});
			WriteElement("key", new string[]{"for", "node", "id", "d5", "attr.name", "url", "attr.type", "string"});
			
			/* TODO:
  			<key for="graphml" id="d4" yfiles.type="resources"/>
  			*/
			
  			WriteStartElement("graph", new string[]{"edgedefault", "directed"});
		}
		
		private void WriteStartElement(string node)
		{
			WriteStartElement(node, null);
		}
		
		private void WriteStartElement(string node, string[] attributes)
		{
			int j = node.IndexOf(":");
			if (j >= 0)
			{
				_writer.WriteStartElement(node.Substring(0, j), node.Substring(j + 1), null);
			}
			else
			{
				_writer.WriteStartElement(node);
			}
			if (attributes != null)
			{
				for(int i = 0; i < attributes.Length; i += 2)
				{
					_writer.WriteAttributeString(attributes[i], attributes[i + 1]);
				}
			}
		}

		private void WriteEndElement()
		{
			_writer.WriteEndElement();
		}
		
		private void WriteElement(string node, string[] attributes)
		{
			WriteStartElement(node, attributes);
			WriteEndElement();
		}
		
		private string EdgeId(int no)
		{
			return "e" + no.ToString();
		}
		
		private string NodeId(int no)
		{
			return "n" + no.ToString();
		}
		
		private static string Encode(string plainText)
		{
			return plainText;
		}
		
		#region IWriter interface
		public override void Write(Edge edge)
		{
			WriteStartElement("edge", new string[]{"id", EdgeId(_edgeId++), "source", NodeId(edge.SourceId), "target", NodeId(edge.TargetId)});
			WriteStartElement("data", new string[]{"key", "d2"});
			WriteStartElement("y:PolyLineEdge");

			if ((edge.Connectors != null) && (edge.Connectors.Length > 0))
			{
				WriteStartElement(
					"y:Path",
					new string[]{
						"sx", "0.0",
						"sy", "0.0",
						"tx", "0.0",
						"ty", "0.0"
					});
				for(int i = 0; i < edge.Connectors.Length; i++)
				{
					WriteElement(
						"y:Point",
						new string[]{
							"x", edge.Connectors[i].X.ToString(),
							"y", edge.Connectors[i].Y.ToString()
						});
				}
				WriteEndElement();
			}
			
			if (!String.IsNullOrEmpty(edge.Text))
			{
				WriteStartElement(
					"y:EdgeLabel",
					new string[]{
						"alignment", "center",
						"backgroundColor", yWorks.Constants.Encode(edge.BackgroundColor),	// TODO: what if edge.BackgroundColor.IsEmpty?
						"fontFamily", edge.Font.Name,
						"fontSize", yWorks.Constants.EncodeFontSize(edge.Font).ToString(System.Globalization.NumberFormatInfo.InvariantInfo),
						"fontStyle", "plain",
						"hasLineColor", "false",
						"modelName", "three_center",
						"modelPosition", "tcentr",
						"textColor", yWorks.Constants.Encode(edge.TextColor),
						"visible", yWorks.Constants.Encode(edge.TextVisible)
				    });
				_writer.WriteString(Encode(edge.Text));
				WriteEndElement();
			}
			
			/* TODO:
          <y:EdgeLabel distance="2.0" height="18.701171875" lineColor="#000000" preferredPlacement="anywhere" ratio="0.5" width="61.0" x="86.52193450927734" y="36.20756530761719">
			
          <y:EdgeLabel distance="2.0" hasLineColor="false" height="16.4501953125" preferredPlacement="anywhere" ratio="0.5" width="10.0" x="1193.9285335348868" y="183.21240234375">
			*/

			if (edge.Width > 0)
			{
				if (edge.Width > 5)
				{
					throw new ArgumentOutOfRangeException();
				}
				WriteElement("y:LineStyle", new string[]{"color", yWorks.Constants.Encode(edge.Color), "type", yWorks.Constants.Encode(edge.Style), "width", edge.Width.ToString()});
			}
			
			WriteElement("y:Arrows", new string[]{"source", "none", "target", yWorks.Constants.Encode(edge.TargetArrow)});
			
			if (edge.SmoothBends)
			{
				WriteElement("y:BendStyle", new string[]{"smoothed", "true"});
			}
			
			WriteEndElement();
			WriteEndElement();

			if (!String.IsNullOrEmpty(edge.Description))
			{
				WriteStartElement("data", new string[]{"key", "d3"});
				_writer.WriteString(Encode(edge.Description));
				WriteEndElement();
			}
			
			if (!String.IsNullOrEmpty(edge.Url))
			{
				WriteStartElement("data", new string[]{"key", "d4"});
				_writer.WriteString(Encode(edge.Url));
				WriteEndElement();
			}
			
			WriteEndElement();
		}
		
		private const int HorizontalSpace = 2; 	// TODO: TUNE:
		private const int VerticalSpace = 2; 	// TODO: TUNE:
		
		public override void Write(Node node)
		{
			WriteStartElement("node", new string[]{"id", NodeId(node.Id)});
			WriteStartElement("data", new string[]{"key", "d0"});
			WriteStartElement("y:ShapeNode");

			if (!String.IsNullOrEmpty(node.Text))
			{
				System.Drawing.Size bounds = Canvas.MeasureString(node.Text, node.Font).ToSize();
				bounds.Width = bounds.Width + 2 * HorizontalSpace;
				bounds.Height = bounds.Height + 2 * VerticalSpace;
				
				WriteElement("y:Geometry", new string[]{
				             	"x", (node.Anchor.X - (bounds.Width / 2)).ToString(),
				             	"y", (node.Anchor.Y - (bounds.Height / 2)).ToString(),
				             	"width", bounds.Width.ToString(),
				             	"height", bounds.Height.ToString()});
				WriteStartElement(
					"y:NodeLabel",
					new string[]{
						"alignment", "center",
						"autoSizePolicy", "content",
						"fontFamily", node.Font.Name,
						"fontSize", yWorks.Constants.EncodeFontSize(node.Font).ToString(System.Globalization.NumberFormatInfo.InvariantInfo),
						"fontStyle", "plain",
						"hasBackgroundColor", "false",
						"hasLineColor", "false",
						"modelName", "internal",
						"modelPosition", "c",
						"textColor", yWorks.Constants.Encode(node.TextColor),
				    });
				_writer.WriteString(Encode(node.Text));
				WriteEndElement();
				/* TODO:
          		  <y:NodeLabel height="16.4501953125" visible="true" width="66.0" x="1.5" y="-1.72509765625">&lt;native&gt;.Any</y:NodeLabel>
				*/			
			}
			else
			{
				System.Drawing.Size bounds = new System.Drawing.Size(10, 10);	// TODO: TUNE: default size
				
				WriteElement("y:Geometry", new string[]{
				             	"x", (node.Anchor.X - (bounds.Width / 2)).ToString(),
				             	"y", (node.Anchor.Y - (bounds.Height / 2)).ToString(),
				             	"width", bounds.Width.ToString(),
				             	"height", bounds.Height.ToString()});
			}
			
			WriteElement("y:Shape", new string[]{"type", yWorks.Constants.Encode(node.Shape)});

			if (node.BackgroundColor.IsEmpty)
			{
				WriteElement("y:Fill", new string[]{"hasColor", "false"});
			}
			else
			{
				WriteElement("y:Fill", new string[]{"color", yWorks.Constants.Encode(node.BackgroundColor)});
			}
			
			if (node.BorderColor.IsEmpty)
			{
				WriteElement("y:BorderStyle", new string[]{"hasColor", "false"});
			}
			else
			{
				WriteElement("y:BorderStyle", new string[]{"color", yWorks.Constants.Encode(node.BorderColor), "type", yWorks.Constants.Encode(node.BorderStyle)});
			}
			
			if ((node.ShadowSize > 0) && !node.ShadowColor.IsEmpty)
			{
				WriteElement("y:DropShadow",
				             new string[]{
				             	"color", yWorks.Constants.Encode(node.ShadowColor),
				             	"offsetX", node.ShadowSize.ToString(),
				             	"offsetY", node.ShadowSize.ToString()
				             });
			}

			WriteEndElement();
			WriteEndElement();
			
			if (!String.IsNullOrEmpty(node.Description))
			{
				WriteStartElement("data", new string[]{"key", "d1"});
				_writer.WriteString(Encode(node.Description));
				WriteEndElement();
			}
			
			if (!String.IsNullOrEmpty(node.Url))
			{
				WriteStartElement("data", new string[]{"key", "d5"});
				_writer.WriteString(Encode(node.Url));
				WriteEndElement();
			}
			
			WriteEndElement();
		}
		
		public override void Close()
		{
			if (_writer != null)
			{
				WriteEndElement();
				WriteEndElement();
				
				_writer.Close();
				_writer = null;
			}
			base.Close();
		}
		#endregion
	}
}
