using System;

namespace Structure.Graph2D.IO
{
	using INodeList = System.Collections.Generic.IEnumerable<Node>;
	using IEdgeList = System.Collections.Generic.IEnumerable<Edge>;
		
	public class Container: WriterBase, IReaderCallback
	{
		private System.Collections.Generic.List<Node> _nodes = new System.Collections.Generic.List<Node>();
		private System.Collections.Generic.List<Edge> _edges = new System.Collections.Generic.List<Edge>();
		
		public class NodeList: INodeList
		{
			private Container _owner;
			
			internal NodeList(Container owner)
			{
				_owner = owner;	
			}
			
			private System.Collections.Generic.List<Node> ListCopy()
			{
				System.Collections.Generic.List<Node> result = new System.Collections.Generic.List<Node>();
				result.AddRange(_owner._nodes);
				return result;
			}
			
			public System.Collections.Generic.IEnumerator<Node> GetEnumerator()
			{
				return ListCopy().GetEnumerator();
			}

			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
			{
				return ListCopy().GetEnumerator();
			}
			
			private bool InternalRemove(Node item)
			{
				return _owner._nodes.Remove(item);
			}
			
			private static bool IsLooping(Edge item)
			{
				return item.SourceId == item.TargetId;
			}
			
			public bool Remove(Node item)
			{
				if (InternalRemove(item))
				{
					// now we need to redirect all edges coming in and going out
					IEdgeList inputs = _owner.Edges.InputsOf(item);
					IEdgeList outputs = _owner.Edges.OutputsOf(item);
				
					foreach(Edge input in inputs)
					{
						if (!IsLooping(input))
						{
							foreach(Edge output in outputs)
							{
								if (!IsLooping(output))
								{
									Edge merged = input.Clone();
									merged.AddConnector(item.Anchor);
									merged.AddConnectors(output.Connectors);
									merged.SourceId = input.SourceId;
									merged.TargetId = output.TargetId;
									_owner.Edges.Add(merged);
								}
							}
						}
					}
					
					_owner.Edges.Remove(inputs);
					_owner.Edges.Remove(outputs);
					
					return true;
				}
				else
				{
					return false;
				}
			}
		}
		
		private NodeList _nodeList;
		public NodeList Nodes
		{
			get
			{
				return _nodeList;
			}
		}
		
		public class EdgeList: IEdgeList
		{
			private Container _owner;
			
			internal EdgeList(Container owner)
			{
				_owner = owner;	
			}
			
			private System.Collections.Generic.List<Edge> ListCopy()
			{
				System.Collections.Generic.List<Edge> result = new System.Collections.Generic.List<Edge>();
				result.AddRange(_owner._edges);
				return result;
			}
			
			public System.Collections.Generic.IEnumerator<Edge> GetEnumerator()
			{
				return ListCopy().GetEnumerator();
			}

			System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
			{
				return ListCopy().GetEnumerator();
			}
			
			private const int AnyId = -1;
			
			private System.Collections.Generic.IEnumerable<Edge> Links(int sourceId, int targetId)
			{
				System.Collections.Generic.List<Edge> result = new System.Collections.Generic.List<Edge>();
				foreach(Edge item in _owner._edges)
				{
					if (((sourceId == AnyId) || (sourceId == item.SourceId)) &&
					    ((targetId == AnyId) || (targetId == item.TargetId)))
					{
						result.Add(item);
					}
				}
				return result;
			}
			
			public System.Collections.Generic.IEnumerable<Edge> InputsOf(Node item)
			{
				return Links(AnyId, item.Id);
			}

			public System.Collections.Generic.IEnumerable<Edge> OutputsOf(Node item)
			{
				return Links(item.Id, AnyId);
			}
			
			public bool Remove(Edge item)
			{
				return _owner._edges.Remove(item);
			}
			
			public void Remove(IEdgeList items)
			{
				foreach(Edge item in items)
				{
					Remove(item);
				}
			}
			
			public void Add(Edge item)
			{
				_owner._edges.Add(item);
			}
		}
		
		private EdgeList _edgeList;
		public EdgeList Edges
		{
			get
			{
				return _edgeList;
			}
		}
		
		public Container()
		{
			_nodeList = new NodeList(this);
			_edgeList = new EdgeList(this);
		}
		
		public void Clear()
		{
			_nodes.Clear();
			_edges.Clear();
		}
		
		#region WriterBase overrides
		public override void Close()
		{
			base.Close();
			Clear();
		}
		
		public override void Write(Edge item)
		{
			_edges.Add(item);
		}
		
		public override void Write(Node item)
		{
			if (item.Id == 0)
			{
				item.Id = _nodes.Count + 1;
			}
			_nodes.Add(item);
		}
		#endregion
		
		#region IReaderCallback implementation
		public INode Add(Node item)
		{
			Write(item);
			return null;
		}
		
		public IEdge Add(Edge item, INode source, INode target)
		{
			Write(item);
			return null;
		}
		#endregion
		
		public void WriteTo(IWriter writer)
		{
			foreach(Node node in _nodes)
			{
				writer.Write(node);
			}
			
			foreach(Edge edge in _edges)
			{
				writer.Write(edge);
			}
		}
	}
}
