using System;
using Structure.Graph2D.IO;

namespace Structure.Graph2D.PaceStar
{
	internal class StringList: System.Collections.Generic.List<string>
	{
		public void LoadFromFile(string fileName)
		{
			System.IO.TextReader reader = System.Text.FileStream.CreateReader(fileName);
			try
			{
				string line;
				while ((line = reader.ReadLine()) != null)
				{
					this.Add(line);
				}
			}
			finally
			{
				reader.Dispose();
			}
		}
	}
	
	public class Reader: ReaderBase
	{
		private static int DiagramToGraph(int coordinate)
		{
			return coordinate / 3 /* TODO: is this heuristic correct? should we determine it from the graph file? */;
		}
		
		private static System.Drawing.Point DiagramToGraph(int x, int y)
		{
			return new System.Drawing.Point(DiagramToGraph(x), DiagramToGraph(y));
		}
		
		private static System.Drawing.Point DiagramToGraph(System.Drawing.Rectangle bounds)
		{
			return DiagramToGraph(
				bounds.X + (bounds.Width / 2),
				bounds.Y + (bounds.Height / 2));
		}
		
		private static string CamelCase(string item)
		{
			return item.Substring(0, 1).ToUpper() + item.Substring(1);
		}
		
		private static string NormalizedId(string code)
		{
			if (String.IsNullOrEmpty(code))
			{
				return "";
			}
			// remove all non-alpha and non-numeric characters and all whitespaces
			// force Camel-casing
			System.Text.StringBuilder buffer;
			buffer = new System.Text.StringBuilder();
			foreach(char c in code.Trim())
			{
				if (Char.IsLetterOrDigit(c))
				{
					buffer.Append(c);
				}
				else
				{
					buffer.Append(' ');
				}
			}
			// reduce consecutive number of spaces to 1
			code = buffer.ToString();
			while (true)
			{
				int i = code.IndexOf("  ");
				if (i < 0)
				{
					break;
				}
				code = code.Remove(i, 1);
			}
			buffer = new System.Text.StringBuilder();
			foreach(string item in code.Split(' '))
			{
				buffer.Append(CamelCase(item));
			}
			return buffer.ToString();
		}
		
		public static void Load(string fileName, IReaderCallback visitor)
		{
			Reader reader = new Reader(visitor);
			try
			{
				StringList lines = new StringList();
				lines.LoadFromFile(fileName);
				reader.Parse(lines);
				foreach(Figure figure in reader._figures.Values)
				{
					if (!FlowCreator.IsJumpPoint(figure))
					{
						Node node = new Node();
						node.Anchor = DiagramToGraph(figure.Bounds);
						node.BackgroundColor = figure.FillColor;
						node.BorderColor = figure.BorderColor;
						node.TextColor = figure.TextColor;
						node.Text = figure.Text;
						node.Url = figure.Link;
						if (figure.Shadow)
						{
							node.ShadowColor = System.Drawing.Color.FromArgb(153, 153, 153);
							node.ShadowSize = DiagramToGraph((reader._shadowX + reader._shadowY) / 2);
						}
						node.Class = NormalizedId(figure.Style);
						if (String.IsNullOrEmpty(figure.Text))
						{
							node.Shape = NodeShape.RoundRectangle;
						}
						node.Font = new System.Drawing.Font(figure.FontName, figure.FontSize * 72 /* Point unit */ / 96 /* DPI */);
						reader.Add(node, figure.ID);
					}
				}

				/* TODO: ?
				foreach(Connector connector in reader._connectors.Values)
				{
					Edge edge = new Edge();
					edge.SmoothBends = true;
					switch (connector.Direction)
					{
						case Connector.ConnectorDirection.FromTo:
							edge.TargetArrow = ArrowStyle.Normal;
							break;
							
						case Connector.ConnectorDirection.None:
							edge.TargetArrow = ArrowStyle.None;
							break;
					}
					reader.Add(edge, connector.FromFigure, connector.ToFigure);
				}
				*/
				foreach(Flow flow in FlowCreator.CreateFlows(reader, null, null))
				{
					Edge edge = new Edge();
					edge.SmoothBends = true;
					edge.TargetArrow = ArrowStyle.Short;
					edge.Class = NormalizedId(flow.Style);
					edge.Color = flow.Color;
					switch (flow.PenStyle)
					{
						case 0:
							edge.Style = EdgeStyle.Normal;
							break;
							
						case 2:
						case 6:
						case 7:
							edge.Style = EdgeStyle.Dotted;
							break;

						case 1:
						case 12:
							edge.Style = EdgeStyle.Dashed;
							break;
							
						default:
							// TOOD: implement
							throw new NotImplementedException(String.Format("PenStyle {0}", flow.PenStyle));
							
					}
					for(int i = 1; i < flow.Connectors.Count; i++)
					{
						edge.AddConnector(DiagramToGraph(reader.GetFigure(flow.Connectors[i].FromFigure).Bounds));
					}
					reader.Add(edge, flow.FromFigure, flow.ToFigure);
				}
			}
			finally
			{
				reader.Close();
			}
		}
		
		#region Data model
		private class FigureStyle
		{
			public System.Drawing.Color FillColor;
			public System.Drawing.Color BorderColor;
			public System.Drawing.Color TextColor;
			public int FontSize;
			public string FontName;
			
			public FigureStyle()
			{
				FillColor = DefaultFillColor;
				BorderColor = DefaultBorderColor;
				TextColor = DefaultTextColor;
			}
		}

		private class FigureStyles: System.Collections.Generic.Dictionary<string, FigureStyle>
		{
			public FigureStyle Get(string id)
			{
				FigureStyle result;
				if (!TryGetValue(id, out result))
				{
					result = null;
				}
				return result;
			}
		}
		
		private FigureStyles _figureStyles = new FigureStyles();

		private class Figure
		{
			public string ID;
			public string Style;
			public string Text;
			public bool HasSymbol;
			public bool IsLabel;
			public System.Drawing.Rectangle Bounds;
			public string Link;
			public System.Drawing.Color FillColor;
			public System.Drawing.Color BorderColor;
			public System.Drawing.Color TextColor;
			public bool Shadow;
			public Labels Labels = new Labels();
			public int FontSize;
			public string FontName;
			
			public Figure()
			{
				Bounds = System.Drawing.Rectangle.Empty;
				FillColor = DefaultFillColor;
				BorderColor = DefaultBorderColor;
				TextColor = DefaultTextColor;
			}
		}
		
		private class Figures: System.Collections.Generic.Dictionary<string, Figure>
		{
			public Figure Get(string id)
			{
				Figure result;
				if (!TryGetValue(id, out result))
				{
					result = null;
				}
				return result;
			}
		}
		
		private Figures _figures = new Figures();
		
		private class Label: Figure
		{
			public enum LabelPosition
			{
				Unknown,
				StartPoint,
				EndPoint
			}
			
			public LabelPosition Position;
			
			public Label(): base()
			{
			}
			
			public Label(Figure source): base()
			{
				this.ID = source.ID;
				this.Style = source.Style;
				this.Text = source.Text;
				this.HasSymbol = source.HasSymbol;
				this.IsLabel = true;
				this.Bounds = source.Bounds;
				this.Link = source.Link;
				this.FillColor = source.FillColor;
				this.BorderColor = source.BorderColor;
				this.TextColor = source.TextColor;
				this.Shadow = source.Shadow;
			}
		}
		
		private class Labels: System.Collections.Generic.Dictionary<string, Label>
		{
			public Label Get(string id)
			{
				Label result;
				if (!TryGetValue(id, out result))
				{
					result = null;
				}
				return result;
			}
			
			public void Add(Label item)
			{
				Add(item.ID, item);
			}
		}
		
		private Labels _labels = new Labels();

		private class Connector
		{
			public enum ConnectorDirection
			{
				None,
				FromTo,
				Both
			}

			public string ID;
			public string Style;
			public string FromFigure;
			public string ToFigure;
			public int PenStyle;
			public ConnectorDirection Direction;
			public System.Drawing.Color Color;
			public Labels Labels = new Labels();
		}

		private class Connectors: System.Collections.Generic.Dictionary<string, Connector>
		{
			public Connector Get(string id)
			{
				Connector result;
				if (!TryGetValue(id, out result))
				{
					result = null;
				}
				return result;
			}
		}
		
		private Connectors _connectors = new Connectors();

		private class ConnectorList: System.Collections.Generic.List<Connector>
		{
		}
		
		private class Staple
		{
			public string ID;
			public int Style;
			public string FromFigure;
			public string ToFigure;
			public string Connector;
		}
		
		private class Staples: System.Collections.Generic.Dictionary<string, Staple>
		{
		}
		
		private Staples _staples = new Staples();
		
		private class Flow
		{
			private class BitFlags
			{
				private string _codes = "";
				
				private static bool Decode(char c)
				{
					return (c == '1');
				}
				
				private static char Encode(bool value)
				{
					return value ? '1': '0';
				}
				
				public bool this[int index]
				{
					get
					{
						return Decode(_codes[index]);
					}
				}
				
				public void Add(bool value)
				{
					_codes = _codes + Encode(value);
				}
				
				public void Insert(bool value)
				{
					_codes = Encode(value) + _codes;
				}
			}
			
			// TODO: public Labels Labels = new Labels();
			public string FromFigure;
			public string ToFigure;
			public ConnectorList Connectors = new ConnectorList();
			private BitFlags _reverseConnectors = new BitFlags();

			public Flow()
			{
			}
			
			public Flow(Flow source)
			{
				this.FromFigure = source.FromFigure;
				this.ToFigure = source.ToFigure;
				for(int i = 0; i < source.Connectors.Count; i++)
				{
					Add(source.Connectors[i], source._reverseConnectors[i]);
				}
			}

			private void Insert(Connector connector, bool reverse)
			{
				_reverseConnectors.Insert(reverse);
				Connectors.Insert(0, connector);
			}
			
			private void Insert(Connector connector, string fromFigure, string toFigure)
			{
				bool reverse = (connector.FromFigure == toFigure) || (connector.ToFigure == fromFigure);
				Insert(connector, reverse);
			}
			
			public void Insert(Connector connector)
			{
				Insert(connector, connector.FromFigure, connector.ToFigure);
				BindEnds();
			}
			
			private void Add(Connector connector, bool reverse)
			{
				_reverseConnectors.Add(reverse);
				Connectors.Add(connector);
			}
			
			private void Add(Connector connector, string fromFigure, string toFigure)
			{
				bool reverse = (connector.FromFigure == toFigure) || (connector.ToFigure == fromFigure);
				Add(connector, reverse);
			}
			
			public void Add(Connector connector)
			{
				Add(connector, connector.FromFigure, connector.ToFigure);
				BindEnds();
			}
			
			private void FetchPoints(int connectorIndex, out string fromFigure, out string toFigure)
			{
				Connector connector = Connectors[connectorIndex];
				if (_reverseConnectors[connectorIndex])
				{
					fromFigure = connector.ToFigure;
					toFigure = connector.FromFigure;
				}
				else
				{
					fromFigure = connector.FromFigure;
					toFigure = connector.ToFigure;
				}
			}
			
			private void BindEnds()
			{
				string dummy;
				FetchPoints(0, out FromFigure, out dummy);
				FetchPoints(Connectors.Count - 1, out dummy, out ToFigure);
			}
			
			public bool Contains(string figureID)
			{
				foreach(Connector connector in Connectors)
				{
					if ((connector.FromFigure == figureID) || (connector.ToFigure == figureID))
					{
						return true;
					}
				}
				return false;
			}

			private bool GetConnector(out Connector connector)
			{
				if (Connectors.Count == 0)
				{
					connector = null;
					return false;
				}
				else
				{
					connector = Connectors[0];
					return true;
				}
			}
			
			public string Style
			{
				get
				{
					Connector connector;
					if (GetConnector(out connector))
					{
						return connector.Style;
					}
					else
					{
						return null;
					}
				}
			}
			
			public int PenStyle
			{
				get
				{
					Connector connector;
					if (GetConnector(out connector))
					{
						return connector.PenStyle;
					}
					else
					{
						return 0;
					}
				}
			}
			
			public System.Drawing.Color Color
			{
				get
				{
					Connector connector;
					if (GetConnector(out connector))
					{
						return connector.Color;
					}
					else
					{
						return Reader.DefaultLineColor;
					}
				}
			}
		}

		private class Flows: System.Collections.Generic.List<Flow>
		{
		}
		
		private FigureStyle GetFigureStyle(string id)
		{
			return _figureStyles.Get(id);
		}
		
		private Figure GetFigure(string id)
		{
			Figure result = _figures.Get(id);
			if (result == null)
			{
				result = _labels.Get(id);
			}
			return result;
		}
		
		private Connector GetConnector(string id)
		{
			return _connectors.Get(id);
		}
		
		private Label GetLabel(string id)
		{
			return _labels.Get(id);
		}
		#endregion
		
		private Reader(IReaderCallback visitor)
			: base(visitor)
		{
		}

		#region Loader's state variables
		private StringList _lines;
		private string _line;
		private int _linePointer;
		private string _name;
		private int _shadowX;
		private int _shadowY;
		#endregion
		
		#region Input parser
		private bool IsSpace(char c)
		{
			switch (c)
			{
				case ' ':
				case '\t':
				case ',':
				case '=':
					return true;
					
				default:
					return false;
			}
  		}
		
		private string CutToken()
		{
			int start = 0;
			int max = _line.Length - 1;

			while (start <= max)
			{
				if (IsSpace(_line[start]))
				{
					start++;
				}
				else
				{
					break;
				}
			}
			
			if (start > max)
			{
				_line = "";
				return "";
			}

			int stop = start + 1;
			
			while (stop <= max)
			{
				if (IsSpace(_line[stop]))
				{
					break;
				}
				else
				{
					stop++;
				}
			}
			
			string result = _line.Substring(start, stop - start);
			_line = _line.Remove(0, stop);
			return result;
		}
		
		private string CutTokenUpperCase()
		{
			return CutToken().ToUpper();
		}

		private string CutTailToken()
		{
			int start = 0;
			int max = _line.Length - 1;
			
			while (start <= max)
			{
				if (IsSpace(_line[start]))
				{
					start++;
				}
				else
				{
					break;
				}
			}
			
			string result = _line.Substring(start);
			_line = "";
			return result;
		}
		
		private bool GetPropName()
		{
			_name = CutTokenUpperCase();
			return !String.IsNullOrEmpty(_name);
		}
		
		private string StringArg()
		{
			string result = CutTailToken();
			if (!String.IsNullOrEmpty(result))
			{
				result = result.Substring(1, result.Length - 2);
			}
			return result;
		}
		
		private string IDArg()
		{
			string result = CutToken();
			if (result == "-1")
			{
				return "";
			}
			else
			{
				return result;
			}
		}
		
		private long NumArg()
		{
			string s = CutTokenUpperCase();
			if (s.StartsWith("0X"))
			{
				return long.Parse(s.Substring(2), System.Globalization.NumberStyles.HexNumber);
			}
			else
			{
				long result;
				if (!long.TryParse(s, out result))
				{
					result = 0;
				}
				return result;
			}
		}
		
		private bool BoolArg()
		{
			string s = CutTokenUpperCase();
			if ((s == "1") || s.StartsWith("T") || s.StartsWith("Y"))
			{
				return true;
			}
			else
			{
				return false;
			}
		}

		private System.Drawing.Color ColorArg()
		{
			int r = (int)NumArg();
			int g = (int)NumArg();
			int b = (int)NumArg();
			if ((r == 1) && (g == 2) && (b == 3))
			{
				return System.Drawing.Color.Empty;
			}
			else
			{
				return System.Drawing.Color.FromArgb(r, g, b);
			}
		}

		private System.Drawing.Rectangle RectArg()
		{
			const int offset1 = 0;	// TODO: should be 1?
			
			int left = (int)NumArg();
			int top = (int)NumArg();
			int right = (int)NumArg();
			int bottom = (int)NumArg();
			return new System.Drawing.Rectangle(left, top, right - left + offset1, bottom - top + offset1);
		}

		private static System.Drawing.Color DefaultFillColor = System.Drawing.Color.White;
		private static System.Drawing.Color DefaultBorderColor = System.Drawing.Color.Black;
		private static System.Drawing.Color DefaultTextColor = System.Drawing.Color.Black;
		private static System.Drawing.Color DefaultLineColor = System.Drawing.Color.Black;
		
		private void LoadAndClear()
		{
			_linePointer += 2;	// throw type and block opening
			int depth = 1;
			while (depth != 0)
			{
				string line = _lines[_linePointer];
				if (line == "{")
				{
					depth++;
				}
				else if (line == "}")
				{
					depth--;
				}
				_linePointer++;
			}
		}
		
		private string TrimLeft(string line)
		{
			return line.TrimStart(new char[]{' ', '\t'});
		}
		
		private string TrimRight(string line)
		{
			return line.TrimEnd(new char[]{' ', '\t'});
		}

		private void LoadFigureStyle()
		{
			FigureStyle style = new FigureStyle();
			_figureStyles.Add(StringArg(), style);
			_linePointer += 2;
			while (true)
			{
				_line = _lines[_linePointer];
				_linePointer++;
				if (_line == "}")
				{
					break;
				}
				else
				{
					if (!GetPropName())
					{
						continue;
					}
					switch (_name)
					{
						case "FILLCOLOR":
							style.FillColor = ColorArg();
							break;
							
						case "BORDERCOLOR":
							style.BorderColor = ColorArg();
							break;
							
						case "TEXTCOLOR":
							style.TextColor = ColorArg();
							break;
							
						case "GRAPHIC":
							if (style.FillColor.IsEmpty)
							{
								style.FillColor = DefaultFillColor;
							}
							break;
							
						case "TYPESIZE":
							style.FontSize = (int)NumArg();
							break;
							
						case "TYPEFACE":
							style.FontName = StringArg();
							break;
					}
				}
			}
		}

		private void LoadFigure()
		{
			Figure figure = new Figure();
			figure.ID = IDArg();
			_linePointer += 2;
			while (true)
			{
				_line = _lines[_linePointer];
				_linePointer++;
				if (_line == "}")
				{
					break;
				}
				else
				{
					if (!GetPropName())
					{
						continue;
					}
					switch (_name)
					{
						case "STYLE":
							{
								figure.Style = StringArg();
								FigureStyle style = GetFigureStyle(figure.Style);
								if (style != null)
								{
									figure.FillColor = style.FillColor;
									figure.BorderColor = style.BorderColor;
									figure.TextColor = style.TextColor;
									figure.FontName = style.FontName;
									figure.FontSize = style.FontSize;
								}
							}
							break;
							
						case "TEXT":
							figure.Text = StringArg().Replace("\\line ", "\n").Replace("\\line", "\n").Replace("\\}", "}").Replace("\\{", "{");
							break;
							
						case "LINK":
							figure.Link = StringArg();
							break;
						
						case "SYMBOL":
							if (StringArg() != "null")
							{
								figure.HasSymbol = true;
							}
							break;
							
						case "LABEL":
							figure.IsLabel = BoolArg();
							break;
							
						case "FILLCOLOR":
							figure.FillColor = ColorArg();
							break;
							
						case "BORDERCOLOR":
							figure.BorderColor = ColorArg();
							break;
							
						case "TEXTCOLOR":
							figure.TextColor = ColorArg();
							break;
							
						case "GRAPHIC":
							if (figure.FillColor.IsEmpty)
							{
								figure.FillColor = DefaultFillColor;
							}
							break;
							
						case "TYPESIZE":
							figure.FontSize = (int)NumArg();
							break;
							
						case "TYPEFACE":
							figure.FontName = StringArg();
							break;
							
						case "BOUNDS":
							figure.Bounds = RectArg();
							break;
							
						case "SHADOW":
							figure.Shadow = BoolArg();
							break;
							
						case "BEHAVIOR":
							if ((NumArg() & 0x40000000) != 0)
							{
								// invisible
								figure.Bounds = System.Drawing.Rectangle.Empty;
							}
							break;
					}
				}
			}
			
			if (figure.IsLabel)
			{
				Label label = new Label(figure);
				_labels.Add(label.ID, label);
			}
			else
			{
				_figures.Add(figure.ID, figure);
			}
		}

		private enum ArrowState
		{
			Unknown,
			Yes,
			No
		}

		private void SetArrow(ref ArrowState arrow, bool enabled)
		{
			if (enabled)
			{
				if (arrow == ArrowState.Unknown)
				{
					arrow = ArrowState.Yes;
				}
			}
			else
			{
				arrow = ArrowState.No;
			}
		}
		
		private void LoadConnector()
		{
			Connector connector = new Connector();
			connector.ID = IDArg();
			_linePointer += 2;
			
			ArrowState fromArrow = ArrowState.Unknown;
			ArrowState toArrow = ArrowState.Unknown;
			
			while (true)
			{
				_line = _lines[_linePointer];
				_linePointer++;
				if (_line == "}")
				{
					break;
				}
				else
				{
					if (!GetPropName())
					{
						continue;
					}
					switch (_name)
					{
						case "STYLE":
							connector.Style = StringArg();
							break;
							
						case "PENSTYLE":
							connector.PenStyle = (int)NumArg();
							break;
							
						case "FIGURE1":
							connector.FromFigure = IDArg();
							break;
						
						case "FIGURE2":
							connector.ToFigure = IDArg();
							break;
							
						case "END1":
							SetArrow(ref fromArrow, StringArg() != "null");
							break;
							
						case "END2":
							SetArrow(ref toArrow, StringArg() != "null");
							break;
							
						case "SUPPRESSEND1":
							if (BoolArg())
							{
								SetArrow(ref fromArrow, false);
							}
							break;
							
						case "SUPPRESSEND2":
							if (BoolArg())
							{
								SetArrow(ref toArrow, false);
							}
							break;
							
						case "COLOR":
							connector.Color = ColorArg();
							break;
					}
				}
			}
			
			if ((fromArrow == ArrowState.Yes) && (toArrow == ArrowState.Yes))
			{
				connector.Direction = Connector.ConnectorDirection.Both;
			}
			else if (((fromArrow == ArrowState.Unknown) || (fromArrow == ArrowState.No)) &&
			         ((toArrow == ArrowState.Unknown) || (toArrow == ArrowState.No)))
			{
				connector.Direction = Connector.ConnectorDirection.None;
			}
			else
			{
				connector.Direction = Connector.ConnectorDirection.FromTo;
			}
			
			_connectors.Add(connector.ID, connector);
		}

		private void LoadStaple()
		{
			Staple staple = new Staple();
			staple.ID = IDArg();
			_linePointer += 2;
			while (true)
			{
				_line = _lines[_linePointer];
				_linePointer++;
				if (_line == "}")
				{
					break;
				}
				else
				{
					if (!GetPropName())
					{
						continue;
					}
					switch (_name)
					{
						case "STAPLETYPE":
							staple.Style = (int)NumArg();
							break;
							
						case "CONNECTOR":
							staple.Connector = IDArg();
							break;
							
						case "FIGURE1":
							staple.FromFigure = IDArg();
							break;
							
						case "FIGURE2":
							staple.ToFigure = IDArg();
							break;
					}
				}
			}
			
			_staples.Add(staple.ID, staple);
			
			switch (staple.Style)
			{
				case 1:
					// connector label
					GetConnector(staple.Connector).Labels.Add(GetLabel(staple.FromFigure));
					break;
					
				case 2:
					// figure label
					GetFigure(staple.FromFigure).Labels.Add(GetLabel(staple.ToFigure));
					break;
					
				case 3:
					{
						Connector connector = GetConnector(staple.Connector);
						Label label = GetLabel(staple.ToFigure);
						
						if (staple.FromFigure == connector.FromFigure)
						{
							label.Position = Label.LabelPosition.StartPoint;
						}
						else if (staple.FromFigure == connector.ToFigure)
						{
							label.Position = Label.LabelPosition.EndPoint;
						}
						
						connector.Labels.Add(label);

					}
					break;
			}
		}
		
		private void Parse(StringList contents)
		{
			const int baseLinePointer = 2;
		
			_lines = contents;
			
			if ((_lines.Count < 2) || (_lines[0] != "EDGE Diagram File"))
			{
				return;
			}

			// merge soft breaks and remove comments and trailing/tailing spaces
			_linePointer = baseLinePointer;
			while (_linePointer < _lines.Count)
			{
				string line = TrimRight(_lines[_linePointer]);
				if (String.IsNullOrEmpty(line))
				{
					_lines[_linePointer] = "";
					_linePointer++;
				}
				else if (line.EndsWith("-"))
				{
					// merge soft break
					_lines[_linePointer] = line.Substring(0, line.Length - 1) + _lines[_linePointer + 1].Trim();
					_lines.RemoveAt(_linePointer + 1);
				}
				else
				{
					// remove comment
					line = TrimLeft(line);
					if (line.StartsWith("#"))
					{
						_lines[_linePointer] = "";
						_linePointer++;
					}
					else
					{
						_lines[_linePointer] = line;
						_linePointer++;
					}
				}
					
			}
			
			_shadowX = 0;
			_shadowY = 0;
			
			// roll and throw out uninteresting features
			_linePointer = baseLinePointer;
			while (_linePointer < _lines.Count - 1)
			{
				_line = _lines[_linePointer];
				if (_lines[_linePointer + 1] == "{")
				{
					// this is a block feature
					// parse type name
					GetPropName();

					switch (_name)
					{
						case "FIGURESTYLE":
							LoadFigureStyle();
							break;
							
						case "CONNECTORSTYLE":
							// TODO: LoadConnectorStyle() = FConnectorStyles.Add(StringArg); ldclr
							LoadAndClear();
							break;
							
						case "FIGURE":
							LoadFigure();
							break;
							
						case "CONNECTOR":
							LoadConnector();
							break;
							
						case "STAPLE":
							LoadStaple();
							break;
							
						default:
							LoadAndClear();
							break;
					}
				}
				else
				{
					_linePointer++;	
					if (!GetPropName())
					{
						continue;
					}
					switch (_name)
					{
						case "SHADOWX":
							_shadowX = (int)NumArg();
							break;
							
						case "SHADOWY":
							_shadowY = (int)NumArg();
							break;
							
						case "TEMPLATE":
							// TODO: FStyle := ChangeFileExt(StringArg, '');
							break;
					}
				}
			}
		}
		#endregion
		
		private class FlowCreator
		{
			private class DependencyMatrix: System.Collections.Generic.Dictionary<string, Object>
			{
				private string Key(string input, string output)
				{
					return input + ";" + output;
				}
				
				public void Set(string input, string output, object value)
				{
					string key = Key(input, output);
					object result;
					if (TryGetValue(key, out result))
					{
						Remove(key);
					}
					Add(key, value);
				}
				
				public object Get(string input, string output)
				{
					string key = Key(input, output);
					object result;
					if (!TryGetValue(key, out result))
					{
						result = null;
					}
					return result;
				}
				
				public StringList SelectInputs(string output)
				{
					StringList result = new StringList();
					foreach(string key in Keys)
					{
						if (key.EndsWith(";" + output))
						{
							result.Add(key.Substring(0, key.Length - (1 + output.Length)));
						}
					}
					return result;
				}
			}
			
			private string[] _figureStyles;
			private string[] _connectorStyles;
			private Reader _owner;

			public static Flows CreateFlows(Reader owner, string[] figureStyles, string[] connectorStyles)
			{
				FlowCreator creator = new FlowCreator(owner, figureStyles, connectorStyles);
				return creator.Execute();
			}
			
			public FlowCreator(Reader owner, string[] figureStyles, string[] connectorStyles)
			{
				_owner = owner;
				
				if (figureStyles == null)
				{
					_figureStyles = new string[]{"*"};
				}
				else
				{
					_figureStyles = figureStyles;
				}
				if (connectorStyles == null)
				{
					_connectorStyles = new string[]{"*"};
				}
				else
				{
					_connectorStyles = connectorStyles;
				}
			}

			private bool Match(string style, string styleMask)
			{
				if (String.IsNullOrEmpty(style))
				{
					return false;
				}
				if ((!String.IsNullOrEmpty(styleMask)) && styleMask.EndsWith("*"))
				{
					int maskLength = styleMask.Length - 1;
					if (style.Length < maskLength)
					{
						return false;
					}
					else
					{
						return style.Substring(0, maskLength) == styleMask.Substring(0, maskLength);
					}
				}
				else
				{
					return style == styleMask;
				}
			}
			
			private bool Match(string style, string[] styles)
			{
				foreach(string styleMask in styles)
				{
					if (Match(style, styleMask))
					{
						return true;
					}
				}
				return false;
			}
			
			private bool IsTerminal(Figure figure)
			{
				return Match(figure.Style, _figureStyles);
			}
			
			private bool IsTerminal(string figureID)
			{
				return IsTerminal(_owner.GetFigure(figureID));
			}

			public static bool IsJumpPoint(Figure figure)
			{
				return String.IsNullOrEmpty(figure.Style) || !figure.HasSymbol;
			}
			
			private bool IsJumpPoint(string figureID)
			{
				return IsJumpPoint(_owner.GetFigure(figureID));
			}
			
			DependencyMatrix _connections;
			
			private void SearchForOutputs(Flows flows, Figure inputFigure, Flow template)
			{
				foreach(string outputID in _connections.SelectInputs(inputFigure.ID))
				{
					if (!template.Contains(outputID))
					{
						Connector connector = (Connector)_connections.Get(outputID, inputFigure.ID);
						if (IsTerminal(outputID))
						{
							Flow flow = new Flow(template);
							flow.Insert(connector);
							flows.Add(flow);
						}
						else if (IsJumpPoint(outputID))
						{
							Flow newTemplate = new Flow(template);
							newTemplate.Insert(connector);
							SearchForOutputs(flows, _owner.GetFigure(outputID), newTemplate);
						}
					}
				}
			}
			
			private void SearchForOutputs(Flows flows, Figure inputFigure)
			{
				foreach(string outputID in _connections.SelectInputs(inputFigure.ID))
				{
					Connector connector = (Connector)_connections.Get(outputID, inputFigure.ID);
					if ((connector.Direction == Connector.ConnectorDirection.FromTo) ||
					    (connector.Direction == Connector.ConnectorDirection.Both))
					{
						if (IsTerminal(outputID))
						{
							Flow flow = new Flow();
							flow.Add(connector);
							flows.Add(flow);
						}
						else if (IsJumpPoint(outputID))
						{
							Flow template = new Flow();
							template.Add(connector);
							SearchForOutputs(flows, _owner.GetFigure(outputID), template);
						}
					}
				}
			}
			
			public Flows Execute()
			{
				_connections = new DependencyMatrix();
				Flows result = new Flows();
				
				foreach(Connector connector in _owner._connectors.Values)
				{
					if (Match(connector.Style, _connectorStyles) &&
					    !String.IsNullOrEmpty(connector.FromFigure) &&
					    !String.IsNullOrEmpty(connector.ToFigure))
					{
						_connections.Set(connector.FromFigure, connector.ToFigure, connector);
						if ((connector.Direction == Connector.ConnectorDirection.None) ||
						    (connector.Direction == Connector.ConnectorDirection.Both))
						{
							_connections.Set(connector.ToFigure, connector.FromFigure, connector);
						}
					}
				}
				
				foreach(Figure figure in _owner._figures.Values)
				{
					if (IsTerminal(figure))
					{
						SearchForOutputs(result, figure);
					}
				}
				
				return result;
			}
		}
	}
}
