using System;
using System.Data;
using System.Collections.Generic;

namespace System.Data.Common
{
	public class DbBridge: IDisposable
	{
		#region Adapters - compatibility limitations
		protected abstract class Adapter
		{
			public enum DataType
			{
				Integer,
				Memo,
				String,
				Float
			}
			
			public virtual string GetDataType(DataType type)
			{
				switch(type)
				{
					case DataType.Integer:
						return "integer";

					case DataType.Memo:
						return "text";

					case DataType.String:
						return "text";
						
					case DataType.Float:
						return "double";
				}
				// TODO:
				return null;
			}
			
			public virtual string GetPrimaryKeyClauseForCreateTable(string tableName, string primaryKeyColumns)
			{
				return "";
			}
			
			public string newLine;
			public int maxStringLength;
			public bool memoFieldSupported;
			public string quotedFieldName;
		}

		private class GenericAdapter: Adapter
		{
			public GenericAdapter()
			{
				newLine = "\r\n";
				maxStringLength = Int32.MaxValue;
				memoFieldSupported = true;
				quotedFieldName = null;
			}
		}

		private class JetAdapter: Adapter
		{
			public JetAdapter()
			{
				newLine = "\r\n";
				maxStringLength = 255;
				memoFieldSupported = true;
				quotedFieldName = "[{0}]";
			}

			public override string GetPrimaryKeyClauseForCreateTable(string tableName, string primaryKeyColumns)
			{
				return String.Format(",constraint pk_{0} primary key({1})", tableName, primaryKeyColumns);
			}
		}

		private class MicrosoftSqlAdapter: Adapter
		{
			public MicrosoftSqlAdapter()
			{
				newLine = "\r\n";
				maxStringLength = Int32.MaxValue;
				memoFieldSupported = true;
				quotedFieldName = "[{0}]";
			}

			public override string GetDataType(DataType type)
			{
				switch(type)
				{
					case DataType.Memo:
						return "ntext";

					case DataType.String:
						return "nvarchar";

					default:
						return base.GetDataType(type);
				}
			}
			
			public override string GetPrimaryKeyClauseForCreateTable(string tableName, string primaryKeyColumns)
			{
				// TODO: implement
				return base.GetPrimaryKeyClauseForCreateTable(tableName, primaryKeyColumns);
			}
		}

		private class ExcelAdapter: Adapter
		{
			public ExcelAdapter()
			{
				newLine = "\n";
				maxStringLength = 255;
				memoFieldSupported = false;
				quotedFieldName = "[{0}]";
			}
		}
		#endregion

		protected DbProviderFactory provider = null;
		protected DbConnection db = null;
		protected Adapter adapter = null;

		public struct AdapterSettings
		{
			public string NewLine;
		}

		public AdapterSettings Configuration
		{
			get
			{
				AdapterSettings result;
				result.NewLine = adapter.newLine;
				return result;
			}
		}

		private void TestConnection(DbProviderFactory aFactory, string connectionString)
		{
			DbConnection aConnection = aFactory.CreateConnection();

			try
			{
				aConnection.ConnectionString = connectionString;
				aConnection.Open();
			}
			catch
			{
				aConnection.Dispose();
				throw;
			}

			provider = aFactory;
			db = aConnection;
		}

		public DbBridge(string dataSourceFile)
		{
			switch(System.IO.Path.GetExtension(dataSourceFile).ToLower())
			{
				case ".mdb":
					{
						System.Data.OleDb.OleDbConnectionStringBuilder builder = new System.Data.OleDb.OleDbConnectionStringBuilder();
						builder.Provider = "Microsoft.Jet.OLEDB.4.0";
						builder.DataSource = dataSourceFile;

						TestConnection(System.Data.OleDb.OleDbFactory.Instance, builder.ConnectionString);

						adapter = new JetAdapter();
					}
					break;

				case ".xls":
					{
						System.Data.Odbc.OdbcConnectionStringBuilder builder = new System.Data.Odbc.OdbcConnectionStringBuilder();
						builder.Driver = "Microsoft Excel Driver (*.xls)";
						builder.Add("DBQ", dataSourceFile);
						builder.Add("ReadOnly", 0);

						TestConnection(System.Data.Odbc.OdbcFactory.Instance, builder.ConnectionString);

						adapter = new ExcelAdapter();
					}
					break;

				case ".udl":
					{
						DbConnectionStringBuilder builder = new DbConnectionStringBuilder();
						System.Configuration.IniFile config = new System.Configuration.IniFile(dataSourceFile);
						try
						{
							// TODO: use some smarter method. ALL lines below the [oledb] section
							// in an UDL file should be used as a connection string
							builder.ConnectionString = "Provider=" + config.ReadString("oledb\\Provider", null);
						}
						finally
						{
							config.Dispose();
						}
						TestConnection(System.Data.OleDb.OleDbFactory.Instance, builder.ConnectionString);

						string provider = ((System.Data.OleDb.OleDbConnection)db).Provider;
						if (provider.StartsWith("Microsoft.Jet", true, System.Globalization.CultureInfo.InvariantCulture))
						{
							adapter = new JetAdapter();
						}
						else if (provider.StartsWith("SQLOLEDB", true, System.Globalization.CultureInfo.InvariantCulture))
						{
							adapter = new MicrosoftSqlAdapter();
						}
						else
						{
							// TODO: distinguish other adapters as well
							adapter = new GenericAdapter();
						}
					}
					break;

				case ".dsn":
					{
						System.Data.Odbc.OdbcConnectionStringBuilder builder = new System.Data.Odbc.OdbcConnectionStringBuilder();
						System.Configuration.IniFile config = new System.Configuration.IniFile(dataSourceFile);
						try
						{
							System.Collections.Specialized.NameValueCollection section = config.ReadSection("ODBC");
							foreach(string key in section.AllKeys)
							{
								builder.Add(key, section[key]);
							}
						}
						finally
						{
							config.Dispose();
						}							
						TestConnection(System.Data.Odbc.OdbcFactory.Instance, builder.ConnectionString);

						string provider = builder.Driver;
						if (provider.StartsWith("Microsoft Excel Driver", true, System.Globalization.CultureInfo.InvariantCulture))
						{
							adapter = new ExcelAdapter();
						}
						else
						{
							// TODO: distinguish other adapters as well
							adapter = new GenericAdapter();
						}
					}
					break;

				default:
					throw new System.Data.DataException("Unsupported data source file");
			}
		}

		#region Implementation of IDisposable
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing)
		{
			if(disposing)
			{
				if (db != null)
				{
					db.Dispose();
				}
				db = null;
			}
		}

		~DbBridge()
		{
			Dispose(false);
		}
		#endregion

		protected object Execute(string sqlCommand)
		{
			DbCommand cmd = db.CreateCommand();
			try
			{
				cmd.CommandText = sqlCommand;
				if (sqlCommand.StartsWith("select ", true, System.Globalization.CultureInfo.InvariantCulture))
				{
					try
					{
						return cmd.ExecuteScalar();
					}
					catch
					{
						// convert invalid "select" statement into NULL value
						return null;
					}
				}
				else
				{
					cmd.ExecuteNonQuery();
					return null;
				}
			}
			finally
			{
				cmd.Dispose();
			}
		}

		public System.Data.DataTable LoadTable(string tableName)
		{
			DbCommand cmd = db.CreateCommand();
			try
			{
				try
				{
					cmd.CommandText = String.Format("select * from {0}", tableName);
					DbDataReader reader = cmd.ExecuteReader();
					try
					{
						System.Data.DataTable result = new System.Data.DataTable();
						try
						{
							result.Load(reader);
							return result;
						}
						catch
						{
							result.Dispose();
							throw;
						}
					}
					finally
					{
						reader.Dispose();
					}
				}
				catch
				{
					return null;
				}
			}
			finally
			{
				cmd.Dispose();
			}
		}

		protected string QuotedFieldName(string fieldName)
		{
			// TODO: use OleDbLiterals to get correct value for quoting field names
			if (adapter.quotedFieldName == null)
			{
				return fieldName;
			}
			else
			{
				return String.Format(adapter.quotedFieldName, fieldName);
			}
		}

		protected DbCommand CreateTable(string tableName, System.Data.DataColumnCollection columns, System.Data.DataColumn[] primaryKey)
		{
			// TODO: implement alter table in order to preserve attributes of external system
			if (Execute(String.Format("select count(*) from {0}", tableName)) != null)
			{
				Execute(String.Format("drop table {0}", tableName));
			}

			string insertColumnNames = "";
			string insertValues = "";
			string createFields = "";
			string primaryKeyColumns = "";

			DbCommand result = db.CreateCommand();
			try
			{
				foreach(System.Data.DataColumn column in columns)
				{
					if (!String.IsNullOrEmpty(insertColumnNames))
					{
						insertColumnNames += ",";
						insertValues += ",";
						createFields += ",";
					}
					
					if (primaryKey != null)
					{
						foreach(System.Data.DataColumn primaryKeyColumn in primaryKey)
						{
							if (primaryKeyColumn.ColumnName == column.ColumnName)
							{
								if (!String.IsNullOrEmpty(primaryKeyColumns))
								{
									primaryKeyColumns += ",";
								}
								primaryKeyColumns += QuotedFieldName(column.ColumnName);
								break;
							}
						}
					}

					insertColumnNames += QuotedFieldName(column.ColumnName);
					insertValues += "?";

					DbParameter param = provider.CreateParameter();
					param.ParameterName = column.ColumnName;
					if (column.DataType == typeof(System.String))
					{
						param.DbType = System.Data.DbType.String;
						param.Size = column.MaxLength;
						if (param.Size > adapter.maxStringLength)
						{
							if (!adapter.memoFieldSupported)
							{
								// TODO: value truncation may occur. provide this information to the user
								param.Size = adapter.maxStringLength;
							}
						}
						if (param.Size > adapter.maxStringLength)
						{
							createFields += String.Format("{0} {1}", QuotedFieldName(column.ColumnName), adapter.GetDataType(Adapter.DataType.Memo));
						}
						else
						{
							createFields += String.Format("{0} {1}({2})", QuotedFieldName(column.ColumnName), adapter.GetDataType(Adapter.DataType.String), param.Size);
						}
					}
					else if ((column.DataType == typeof(System.Single)) || (column.DataType == typeof(System.Double)))
					{
						param.DbType = System.Data.DbType.Double;
						createFields += String.Format("{0} {1}", QuotedFieldName(column.ColumnName), adapter.GetDataType(Adapter.DataType.Float));
					}
					else
					{	
						// TODO: map also other data types
						param.DbType = System.Data.DbType.Int32;
						createFields += String.Format("{0} {1}", QuotedFieldName(column.ColumnName), adapter.GetDataType(Adapter.DataType.Integer));
					}
					param.IsNullable = true;
					result.Parameters.Add(param);
				}

				if (String.IsNullOrEmpty(primaryKeyColumns))
				{
					Execute(String.Format("create table {0} ({1})", tableName, createFields));
				}
				else
				{
					Execute(String.Format("create table {0} ({1} {2})", tableName, createFields, adapter.GetPrimaryKeyClauseForCreateTable(tableName, primaryKeyColumns)));
				}

				result.CommandText = String.Format("insert into {0}({1}) values ({2})", tableName, insertColumnNames, insertValues);
				result.Prepare();
			}
			catch
			{
				result.Dispose();
				throw;
			}
			return result;
		}

		private void UpdateTableColumns(System.Data.DataTable table)
		{
			List<System.Data.DataColumn> columns = new List<System.Data.DataColumn>();
			List<int> maxLengths = new List<int>();

			foreach(DataColumn column in table.Columns)
			{
				if ((column.DataType == typeof(System.String)) && (column.MaxLength <= 0))
				{
					columns.Add(column);
					maxLengths.Add(1);
				}
			}

			if (columns.Count == 0) return;

			foreach(DataRow row in table.Rows)
			{
				for(int i = 0; i < columns.Count; i++)
				{
					DataColumn column = columns[i];
					if (!row.IsNull(column))
					{
						string value = row[column].ToString();
						// TODO: convert newLines to database new lines
						int len = value.Length;
						if (len > maxLengths[i])
						{
							maxLengths[i] = len;
						}
					}
				}
			}

			for(int i = 0; i < columns.Count; i++)
			{
				columns[i].MaxLength = maxLengths[i];
			}
		}

		public void CreateTable(string tableName, System.Data.DataTable sourceTable)
		{
			CreateTable(tableName, sourceTable, null, null);
		}
		
		public void CreateTable(string tableName, System.Data.DataTable sourceTable, System.Data.IDataReader rowReader, System.Threading.IProgress observer)
		{
			bool useObserver = observer != null;
			bool useRowReader = rowReader != null;
			if (useObserver)
			{
				if (useRowReader)
				{
					// we don't know the row count
				}
				else
				{
					observer.Open(0, (ulong)sourceTable.Rows.Count);
				}
			}
			try
			{
				UpdateTableColumns(sourceTable);
				DbCommand filler = CreateTable(tableName, sourceTable.Columns, sourceTable.PrimaryKey);
				try
				{
					ulong count = 0;
					if (useRowReader)
					{
						while (rowReader.Read())
						{
							for(int columnIndex = 0; columnIndex < sourceTable.Columns.Count; columnIndex++)
							{
								DbParameter parameter = filler.Parameters[sourceTable.Columns[columnIndex].ColumnName];
		
								if (rowReader.IsDBNull(columnIndex))
								{
									parameter.Value = DBNull.Value;
								}
								else
								{
									string value = rowReader.GetValue(columnIndex).ToString();
									if (parameter.DbType == System.Data.DbType.String)
									{
										int len = value.Length;
										if (len > parameter.Size)
										{
											// value truncation occurs
											// TODO: notify user
											value = value.Substring(0, parameter.Size - 3) + "...";
										}
									}
									parameter.Value = value;
								}
							}
							filler.ExecuteNonQuery();
							if (useObserver)
							{
								observer.Step(count);
								count++;
							}
						}
					}
					else
					{
						foreach(System.Data.DataRow row in sourceTable.Rows)
						{
							foreach(System.Data.DataColumn column in sourceTable.Columns)
							{
								DbParameter parameter = filler.Parameters[column.ColumnName];
		
								if (row.IsNull(column))
								{
									parameter.Value = DBNull.Value;
								}
								else
								{
									string value = row[column].ToString();
									if (parameter.DbType == System.Data.DbType.String)
									{
										int len = value.Length;
										if (len > parameter.Size)
										{
											// value truncation occurs
											// TODO: notify user
											value = value.Substring(0, parameter.Size - 3) + "...";
										}
									}
									parameter.Value = value;
								}
							}
							filler.ExecuteNonQuery();
							if (useObserver)
							{
								observer.Step(count);
								count++;
							}
						}
					}
				}
				finally
				{
					filler.Dispose();
				}
			}
			finally
			{
				if (useObserver)
				{
					observer.Close();
				}
			}
		}
	}
}
