/* KEYWORDS: VERSION, TODO
 * VERSION: 1.6
 * 1.6 / 2008-11-06
 * - new method Csv.CreateWriter() -> IDataWriter
 * 1.5 / 2007-01-02
 * - changed code page in CsvWriter to UTF8 because of compatibility with Excel
 * 1.4 / 2006-12-26
 * - changed code page in CsvReader to autodetect
 * - changed code page in CsvWriter to Unicode (see Csv.DefaultCodePage)
 * 1.3 / 2006-05-29
 * - Fixed: CsvWriter support for multi-line records is Excel compatible
 * 1.2 / 2006-05-20
 * - CsvReader supports reading of multi-line records
 * - CsvReader automaticaly detects separator
 * - CsvReader correctly detects embedded quotes
 * - CsvWriter supports multi-line records
 */
using System;
using System.IO;
using System.Data;

namespace System.Data
{
	public class Csv
	{
		public const char Quote = '"';  // TODO: support other values for quotes

		public static int DefaultCodePage = System.Text.Encoding.UTF8.CodePage; // System.Text.Encoding.Unicode.CodePage

		public static IDataReader ExecuteReader(string fileName, int codepage)
		{
			return ExecuteReader(fileName, codepage, null);
		}
		
		public static IDataReader ExecuteReader(string fileName, int codepage, System.Data.DataColumnCollection schema)
		{
			TextReader stream = null;
			try
			{
				stream = System.Text.FileStream.CreateReader(fileName, codepage);
			}
			catch
			{
				stream = null;
			}
			CsvDataReader result = new CsvDataReader(stream);
			try
			{
				result.CheckSchema(schema);
			}
			catch(Exception)
			{
				result.Dispose();
				throw;
			}
			return result;
		}

		public static IDataReader ExecuteReader(string fileName)
		{
			return ExecuteReader(fileName, DefaultCodePage);
		}

		public static IDataReader ExecuteReader(System.IO.TextReader stream)
		{
			return new CsvDataReader(stream, false);
		}
		
		private static DataTable InternalLoadIntoExistingTableAndCloseReader(DataTable t, IDataReader r)
		{
			try
			{
				if(r.Read())
				{
					do
					{
						DataRow row = t.NewRow();
						for(int i = 0; i < r.FieldCount; i++)
						{
							object value = r.GetValue(i);
							if(value != null)
							{
								row[i] = value;
							}
						}
						t.Rows.Add(row);
						row.AcceptChanges();
					}
					while(r.Read());

					return t;
				}
				else
				{
					return null;
				}
			}
			finally
			{
				r.Close();
			}
		}
		
		private static DataTable InternalLoadTableAndCloseReader(IDataReader r)
		{
			try
			{
				if(r.Read())
				{
					DataTable t = new DataTable();

					for(int i = 0; i < r.FieldCount; i++)
					{
						DataColumn c = new DataColumn();
						c.ColumnName = r.GetName(i);
						c.DataType = r.GetFieldType(i);
						t.Columns.Add(c);
					}

					do
					{
						DataRow row = t.NewRow();
						for(int i = 0; i < r.FieldCount; i++)
						{
							object value = r.GetValue(i);
							if(value != null)
							{
								row[i] = value;
							}
						}
						t.Rows.Add(row);
						row.AcceptChanges();
					}
					while(r.Read());

					return t;
				}
				else
				{
					return null;
				}
			}
			finally
			{
				r.Close();
			}
		}
		
		public static DataTable LoadTable(DataTable table, string fileName, int codepage)
		{
			return InternalLoadIntoExistingTableAndCloseReader(table, ExecuteReader(fileName, codepage, table.Columns));
		}
		
		public static DataTable LoadTable(DataTable table, string fileName)
		{
			return LoadTable(table, fileName, DefaultCodePage);
		}
			
		public static DataTable LoadTable(string fileName, int codepage)
		{
			return InternalLoadTableAndCloseReader(ExecuteReader(fileName, codepage));
		}

		public static DataTable LoadTable(string fileName)
		{
			return LoadTable(fileName, DefaultCodePage);
		}
		
		public static DataTable LoadTable(System.IO.TextReader stream)
		{
			return InternalLoadTableAndCloseReader(ExecuteReader(stream));
		}

		public static System.Data.IDataWriter CreateWriter(string fileName, System.Data.DataColumnCollection schema)
		{
			return new CsvWriter(fileName, DefaultCodePage, schema);
		}
		
		public static System.Data.IDataReader CreateReader(string fileName, System.Data.DataColumnCollection schema)
		{
			return CreateReader(fileName, DefaultCodePage, schema);
		}
		
		public static System.Data.IDataReader CreateReader(string fileName, int codepage, System.Data.DataColumnCollection schema)
		{
			return ExecuteReader(fileName, codepage, schema);
		}
			
		public static System.Data.IDataWriter CreateWriter(string fileName, int codepage, System.Data.DataColumnCollection schema)
		{
			return new CsvWriter(fileName, codepage, schema);
		}
		
		private class CsvWriter: System.Data.IDataWriter
		{
			private StreamWriter pStream = null;
			private bool pNewLine = true;
			private System.Data.DataColumnCollection pSchema = null;
			private string Separator = System.Globalization.CultureInfo.CurrentCulture.TextInfo.ListSeparator;

			public CsvWriter(string fileName, int codepage, System.Data.DataColumnCollection schema)
			{
				pStream = new StreamWriter(fileName, false, System.Text.Encoding.GetEncoding(codepage));
				pSchema = schema;
				if (pSchema != null)
				{
					foreach(DataColumn c in pSchema)
					{
						Write(c.ColumnName /*, true */);
					}
					WriteLine();
				}
			}

			public void Close()
			{
				if(pStream != null)
				{
					pStream.Close();
					pStream = null;
				}
			}

			public void Write(string item, bool forceSeparator)
			{
				if(!pNewLine)
				{
					pStream.Write(Separator);
				}

				if(item != null)
				{
					if(forceSeparator || (item == "") || (item.IndexOfAny(new char[]{' ', '\n', '\r', '\t', Csv.Quote}) >= 0) || (item.IndexOf(Separator) >= 0))
					{
						pStream.Write(Csv.Quote);
						System.Text.StringBuilder replacer = new System.Text.StringBuilder();
						for(int i = 0; i < item.Length; i++)
						{
							if(item[i] == Csv.Quote)
							{
								replacer.Append("\"\"");
							}
							else if(item[i] == '\r')
							{
								// skip
							}
							else if(item[i] == '\n')
							{
								// TODO: use OS specific line terminator
								replacer.Append("\r\n");
							}
							else
							{
								replacer.Append(item[i]);
							}
						}
						pStream.Write(replacer.ToString());
						pStream.Write(Csv.Quote);
					}
					else
					{
						pStream.Write(item);
					}
				}
				pNewLine = false;
			}

			public void Write(string item)
			{
				Write(item, false);
			}

			public void Write(object item)
			{
				if(item == null)
				{
					Write((string)null);
				}
				else
				{
					Write(item.ToString());
				}
			}

			public void WriteLine()
			{
				pStream.WriteLine("");
				pNewLine = true;
			}
			
			#region IDataWriter interface
			public void Write(DataRow row)
			{
				foreach(DataColumn c in pSchema)
				{
					if(row.IsNull(c))
					{
						Write((string)null);
					}
					else
					{
						Write(row[c]);
					}
				}
				WriteLine();
			}
			#endregion
			
			#region IDisposable interface
			public void Dispose()
			{
				Dispose(true);
				GC.SuppressFinalize(this);
			}
			
			private void Dispose(bool commit)
			{
				Close();
			}
			
			~CsvWriter()
			{
				Dispose(false);
			}
			#endregion
		}

		public static void SaveTable(DataTable table, string fileName, int codepage)
		{
			CsvWriter writer = new CsvWriter(fileName, codepage, table.Columns);
			try
			{
				foreach(DataRow r in table.Rows)
				{
					writer.Write(r);
				}
			}
			finally
			{
				writer.Close();
			}
		}

		public static void SaveTable(DataTable table, string fileName)
		{
			SaveTable(table, fileName, DefaultCodePage);
		}
	}
}

namespace System.Data
{
	using System.Collections;
	using System.Collections.Specialized;
	using System.Text;

	class CsvDataReader: IDataReader
	{
		private class ColumnCollection: ArrayList
		{
			public int IndexOf(string name)
			{
				for (int i = 0; i < Count; i++)
				{
					if(this[i].ColumnName == name) return i;
				}
				string u = name.ToUpper();
				for(int i = 0; i < Count; i++)
				{
					if(this[i].ColumnName.ToUpper() == u) return i;
				}
				return -1;
			}

			public new DataColumn this[int i]
			{
				get
				{
					return (DataColumn)base[i];
				}
			}
		}

		private TextReader pStream = null;
		private ColumnCollection pColumns = null;
		private object[] pValues = null;
		private bool pOwnedStream = true;

		public CsvDataReader(TextReader aStream)
		{
			pStream = aStream;
		}
		
		public CsvDataReader(TextReader aStream, bool anOwnedStream)
		{
			pStream = aStream;
			pOwnedStream = anOwnedStream;
		}

		public void CheckSchema(System.Data.DataColumnCollection schema)
		{
			if (schema == null)
			{
				return;
			}
			if (!ReadColumns())
			{
				throw new System.IO.EndOfStreamException();
			}
			for(int i = 0; i < FieldCount; i++)
			{
				if (schema[i].ColumnName != GetName(i))
				{
					throw new Exception("Table format does not match");
				}
			}
		}
		
		#region Implementation of IDataReader
		public virtual void Close()
		{
			if(!IsClosed)
			{
				if (pOwnedStream)
				{
					pStream.Close();
				}
				pStream = null;
				pColumns = null;
				pValues = null;
			}
		}

		public virtual bool NextResult()
		{
			return false;
		}

		#region Separator guessing
		private bool pSeparatorKnown = false;
		private char pSeparator;

		/// <summary>
		/// <returns>value of the separator character or null character if separator is not yet known</returns>
		/// </summary>
		private char GuessedSeparator
		{
			get
			{
				return pSeparatorKnown ? pSeparator : Char.MinValue;
			}
		}

		private char GuessSeparator(string line)
		{
			if(pSeparatorKnown) return pSeparator;

			int pos = 0;

			bool inQuotes = false;

			while(pos < line.Length)
			{
				switch(line[pos])
				{
					case Csv.Quote:
						if(inQuotes)
						{
							pos++;
							if(pos < line.Length)
							{
								if(line[pos] == Csv.Quote)
								{
									pos++;
								}
								else
								{
									inQuotes = false;

									pSeparatorKnown = true;
									pSeparator = line[pos];
									return pSeparator;
								}
							}
							else
							{
								inQuotes = false;
							}
						}
						else
						{
							inQuotes = true;
							pos++;
						}
						break;

					default:
						if(inQuotes)
						{
							pos++;
						}
						// TODO: write the IsIdentifier() detector in a shorter way
						else if("0123456789_.abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ".IndexOf(line[pos]) < 0)
						{
							pSeparatorKnown = true;
							pSeparator = line[pos];
							return pSeparator;
						}
						else
						{
							pos++;
						}
						break;
				}
			}

			// TODO: guessing didn't find anything
			pSeparatorKnown = true;
			pSeparator = Char.MinValue;
			return pSeparator;
		}
		#endregion

		private string[] Split(string line)
		{
			int pos = 0;
			StringCollection buffer = new StringCollection();

			string token = "";
			bool tokenValid = false;
			bool inQuotes = false;
			char Separator = GuessSeparator(line);

			while(pos < line.Length)
			{
				if(line[pos] == Separator)
				{
					if(inQuotes)
					{
						token = token + line[pos];
					}
					else
					{
						buffer.Add(token);
						token = "";
						tokenValid = true;
					}
					pos++;
				}
				else if(line[pos] == Csv.Quote)
				{
					if(inQuotes)
					{
						token = token + line[pos];
						pos++;
						if(pos < line.Length)
						{
							if(line[pos] == Csv.Quote)
							{
								pos++;
							}
							else
							{
								inQuotes = false;
							}
						}
						else
						{
							inQuotes = false;
						}
					}
					else
					{
						if((pos == 0) || (line[pos - 1] == Separator))
						{
							inQuotes = true;
						}
						token = token + line[pos];
						pos++;
					}
				}
				else
				{
					token = token + line[pos];
					pos++;
				}
			}

			if(tokenValid || (token.Length > 0))
			{
				buffer.Add(token);
			}

			string[] result = new string[buffer.Count];
			for(int i = 0; i < buffer.Count; i++)
			{
				result[i] = buffer[i];
			}
			return result;
		}

		private string ReadMultiLine()
		{
			string line = pStream.ReadLine();
			if(line == null) return null;

			string result = line;

			bool inQuotes = false;
			int pos = 0;

		loop:
			while(pos < line.Length)
			{
				if(line[pos] == Csv.Quote)
				{
					if(inQuotes)
					{
						pos++;
						if(pos < line.Length)
						{
							if(line[pos] == Csv.Quote)
							{
								pos++;
							}
							else
							{
								inQuotes = false;
							}
						}
						else
						{
							inQuotes = false;
						}
					}
					else
					{
						if((pos == 0) || (line[pos - 1] == GuessedSeparator))
						{
							inQuotes = true;
						}
						pos++;
					}
				}
				else
				{
					pos++;
				}
			}

			if(inQuotes)
			{
				line = pStream.ReadLine();
				if(line == null)
				{
					// TODO: opened quotes error
					return result;
				}
				result = result + '\n' + line;
				pos = 0;
				goto loop;
			}

			return result;
		}

		private object[] ReadLine(int count)
		{
			string line = ReadMultiLine();
			if(line == null) return null;

			string[] lines = Split(line);
			object[] result;

			if(count < 0)
			{
				count = lines.Length;
			}
			result = new object[count];
			for(int i = 0; i < count; i++)
				result[i] = null;
			if(lines.Length < count)
				count = lines.Length;

			for(int i = 0; i < count; i++)
			{
				string s = lines[i];
				if(s != "")
				{
					string q = new string(Csv.Quote, 1);

					if(s.StartsWith(q))
					{
						s = s.Remove(0, 1);
						if(s.EndsWith(q))
						{
							s = s.Remove(s.Length - 1, 1);
						}
					}
					result[i] = s;
				}
			}

			return result;
		}

		private object[] ReadLine()
		{
			return ReadLine(-1);
		}

		private bool ReadColumns()
		{
			if (pColumns == null)
			{
				object[] header = ReadLine();
				if(header == null)
				{
					return false;
				}
				pColumns = new ColumnCollection();
				for(int i = 0; i < header.Length; i++)
				{
					DataColumn c = new DataColumn();
					c.ColumnName = (string)header[i];
					c.DataType = typeof(string);  // TODO: load schema if available
					pColumns.Add(c);
				}
				return true;
			}
			else
			{
				return true;
			}
		}
		
		public virtual bool Read()
		{
			if(IsClosed) return false;
			if(pColumns == null)
			{
				if (!ReadColumns())
				{
					Close();
					return false;
				}
			}
			pValues = ReadLine(pColumns.Count);
			return (pValues != null);
		}

		public virtual DataTable GetSchemaTable()
		{
			// TODO: implement
			throw new NotImplementedException();
		}

		public int Depth
		{
			get
			{
				return 0;
			}
		}

		public int RecordsAffected
		{
			get
			{
				return -1;
			}
		}

		public bool IsClosed
		{
			get
			{
				return pStream == null;
			}
		}
		#endregion

		#region Implementation of IDataRecord
		private bool ValidColumn(int i)
		{
			return !((IsClosed || (i < 0) || (i >= pColumns.Count)));
		}

		public string GetName(int i)
		{
			return ValidColumn(i) ? pColumns[i].ColumnName : "";
		}

		public string GetDataTypeName(int i)
		{
			return ValidColumn(i) ? pColumns[i].DataType.Name : "";
		}

		public Type GetFieldType(int i)
		{
			return ValidColumn(i) ? pColumns[i].DataType : null;
		}

		public object GetValue(int i)
		{
			return ValidColumn(i) ? pValues[i] : null;
		}

		public int GetValues(object[] values)
		{
			int i = 0, j = 0;
			for (; i < values.Length && j < pValues.Length; i++, j++)
			{
				values[i] = pValues[j];
			}

			return i;
		}

		public int GetOrdinal(string name)
		{
			int index = pColumns.IndexOf(name);
			if (index < 0)
				throw new IndexOutOfRangeException();
			return index;
		}

		public bool GetBoolean(int i)
		{
			return (bool)GetValue(i);
		}

		public byte GetByte(int i)
		{
			return (byte)GetValue(i);
		}

		public long GetBytes(
							 int i,
							 long fieldOffset,
							 byte[] buffer,
							 int bufferoffset,
							 int length)
		{
			throw new NotImplementedException();
		}

		public char GetChar(int i)
		{
			return (char)GetValue(i);
		}

		public long GetChars(
							 int i,
							 long fieldoffset,
							 char[] buffer,
							 int bufferoffset,
							 int length)
		{
			throw new NotImplementedException();
		}

		public IDataReader GetData(int i)
		{
			throw new NotImplementedException();
		}

		public DateTime GetDateTime(int i)
		{
			return (DateTime)GetValue(i);
		}

		public Guid GetGuid(int i)
		{
			return (Guid)GetValue(i);
		}

		public Int16 GetInt16(int i)
		{
			return (Int16)GetValue(i);
		}

		public Int32 GetInt32(int i)
		{
			return (Int32)GetValue(i);
		}

		public long GetInt64(int i)
		{
			return (long)GetValue(i);
		}

		public float GetFloat(int i)
		{
			return (float)GetValue(i);
		}

		public double GetDouble(int i)
		{
			return (double)GetValue(i);
		}

		public string GetString(int i)
		{
			return (string)GetValue(i);
		}

		public decimal GetDecimal(int i)
		{
			return (decimal)GetValue(i);
		}

		public bool IsDBNull(int i)
		{
			return ValidColumn(i) ? (pValues[i] == null) : true;
		}

		public int FieldCount
		{
			get
			{
				return IsClosed ? -1 : pColumns.Count;
			}
		}

		public object this[int i]
		{
			get
			{
				return GetValue(i);
			}
		}

		public object this[string name]
		{
			get
			{
				return GetValue(GetOrdinal(name));
			}
		}
		#endregion

		#region Implementation of IDisposable
		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing)
		{
			if(disposing)
			{
				if(!IsClosed)
				{
					Close();
				}
			}
		}

		~CsvDataReader()
		{
			Dispose(false);
		}
		#endregion
	}
}

