using System;

namespace System.IO
{
	using RecordId = UInt32;
	
	public class PersistentList: System.Collections.Generic.IEnumerable<RecordId>
	{
		private class Enumerator: System.Collections.Generic.IEnumerator<RecordId>
		{
			private RecordId _current;
			private PersistentList _container;
			
			internal Enumerator(PersistentList container)
			{
				_container = container;
				Reset();
			}
			
			public RecordId Current 
			{
				get 
				{
					return _current;
				}
			}
			
			object System.Collections.IEnumerator.Current 
			{
				get 
				{
					return _current;
				}
			}
			
			public void Dispose()
			{
				_container.UnLockRecord(_current);
				_current = NoRecord;
			}
			
			public bool MoveNext()
			{
				_container.Connect();
				try
				{
					RecordId next;
					do
					{
						if (_current == NoRecord)
						{
							next = _container._header.FirstDataRecord;
							if (next == NoRecord)
							{
								return false;
							}
							else
							{
								_container.UnLockRecord(_current);
								_current = next;
								_container.LockRecord(_current);
							}
						}
						else
						{
							Record buffer = _container.ReadRecord(_current, false);
							next = buffer.NextRecordNumber;
							if (next == NoRecord)
							{
								return false;
							}
							else
							{
								_container.UnLockRecord(_current);
								_current = next;
								_container.LockRecord(_current);
							}
						}
					}
					while (_container.IsDeleted(_current));
					return true;
				}
				finally
				{
					_container.Disconnect();
				}
			}
			
			public void Reset()
			{
				_container.UnLockRecord(_current);
				_current = NoRecord;
			}
		}

		public System.Collections.Generic.IEnumerator<RecordId> GetEnumerator()
		{
			return new Enumerator(this);
		}
		
		System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
		{
			return new Enumerator(this);
		}
		
		private System.IO.Stream _stream = null;
		private string _fileName;
		private uint _maxDataSize;

		private const RecordId NoRecord = 0;
		private uint HeaderSignature
		{
			get
			{
				return 0xA3B00000 | _maxDataSize;
			}
		}
		
		private class RecordFlags
		{
			public const uint Deleted				= 1 << 0;
			public const uint ContainsStringData 	= 1 << 1;
			public const uint Locked				= 1 << 2;
			
			public static uint GetLockCount(uint flags)
			{
				return (flags >> 8) & 0xFFFF;
			}
			
			public static void SetLockCount(ref uint flags, uint value)
			{
				flags = (flags & 0xFF0000FF) | (value << 8);
			}
			
			public static void SetBit(ref uint flags, uint bitMask, bool value)
			{
				if (value)
				{
					flags = flags | bitMask;
				}
				else
				{
					flags = flags & ~bitMask;
				}
			}
			
			public static bool GetBit(uint flags, uint bitMask)
			{
				if ((flags & bitMask) == 0)
				{
					return false;
				}
				else
				{
					return true;
				}
			}
		}
		
		private struct Header
		{
			public uint Signature;
			public uint RecordCount;
			public RecordId FirstDataRecord;
			public RecordId LastDataRecord;
			public RecordId FirstFreeRecord;
			public RecordId LastFreeRecord;
		}
		
		private const uint SizeOfHeader = sizeof(uint) + sizeof(uint) + sizeof(RecordId) + sizeof(RecordId) + sizeof(RecordId) + sizeof(RecordId);
		
		private struct Record
		{
			public RecordId RecordNumber;
			
			public RecordId PreviousRecordNumber;
			public RecordId NextRecordNumber;
			public uint Flags;
			public uint DataSize;
			public byte[] Data;
		}
		
		private uint SizeOfRecord
		{
			get
			{
				return /*sizeof(RecordId) + */sizeof(RecordId) + sizeof(RecordId) + sizeof(uint) + sizeof(uint) + _maxDataSize;
			}
		}
		
		private Header _header = new Header();
		
		public PersistentList(string fileName, uint maxDataSize)
		{
			_fileName = System.IO.Path.GetFullPath(fileName);
			_maxDataSize = maxDataSize;
		}

		private void SeekToHeader()
		{
			_stream.Position = 0;
		}
		
		private void SeekToRecord(RecordId recordNumber)
		{
			long position = (long)(SizeOfHeader + ((long)(recordNumber - 1)) * SizeOfRecord);
			if (position > _stream.Length)
			{
				_stream.SetLength(position);
			}
			_stream.Position = position;
		}
		
		private Record ReadRecord(RecordId recordNumber, bool readData)
		{
			Record result = new Record();
			if (recordNumber != NoRecord)
			{
				SeekToRecord(recordNumber);
				result.RecordNumber = recordNumber;
				Read(out result.PreviousRecordNumber);
				Read(out result.NextRecordNumber);
				Read(out result.Flags);
				Read(out result.DataSize);
				if (readData && (result.DataSize > 0))
				{
					result.Data = new byte[result.DataSize];
					Read(result.Data, result.DataSize);
				}
			}
			return result;
		}

		private bool IsDeleted(RecordId recordNumber)
		{
			if (recordNumber == NoRecord)
			{
				return true;
			}
			Record buffer = ReadRecord(recordNumber, false);
			return RecordFlags.GetBit(buffer.Flags, RecordFlags.Deleted);
		}
		
		private void LockRecord(RecordId recordNumber)
		{
			if (recordNumber == NoRecord)
			{
				return;
			}
			Connect();
			try
			{
				Record buffer = ReadRecord(recordNumber, false);
				RecordFlags.SetLockCount(ref buffer.Flags, RecordFlags.GetLockCount(buffer.Flags) + 1);
				RecordFlags.SetBit(ref buffer.Flags, RecordFlags.Locked, true);
				WriteRecord(ref buffer, false);
			}
			finally
			{
				Disconnect();
			}
		}
		
		private void UnLockRecord(RecordId recordNumber)
		{
			if (recordNumber == NoRecord)
			{
				return;
			}
			Connect();
			try
			{
				Record buffer = ReadRecord(recordNumber, false);
				if (RecordFlags.GetBit(buffer.Flags, RecordFlags.Locked))
				{
					uint locks = RecordFlags.GetLockCount(buffer.Flags) - 1;
					RecordFlags.SetLockCount(ref buffer.Flags, locks);
					if (locks == 0)
					{
						RecordFlags.SetBit(ref buffer.Flags, RecordFlags.Locked, false);
						if (RecordFlags.GetBit(buffer.Flags, RecordFlags.Deleted))
						{
							// the record was deleted while being locked. delete it really now
							
							// remove record from the data chain
							InternalRemoveFromChain(ref buffer, ref _header.FirstDataRecord, ref _header.LastDataRecord);
							
							// insert record to the free chain
							InternalAddToChain(ref buffer, ref _header.FirstFreeRecord, ref _header.LastFreeRecord);
						}
					}
					WriteRecord(ref buffer, false);
				}
			}
			finally
			{
				Disconnect();
			}
		}
		
		private void WriteRecord(ref Record buffer, bool writeData)
		{
			if (buffer.RecordNumber != NoRecord)
			{
				SeekToRecord(buffer.RecordNumber);
				Write(buffer.PreviousRecordNumber);
				Write(buffer.NextRecordNumber);
				Write(buffer.Flags);
				if (writeData)
				{
					Write(buffer.DataSize);
					if (buffer.DataSize > 0)
					{
						Write(buffer.Data, buffer.DataSize);
					}
				}
			}
		}

		public RecordId Add(byte[] data)
		{
			return InternalAdd(data, 0);
		}
		
		public RecordId Add(string data)
		{
			if (data == null)
			{
				return InternalAdd(null, RecordFlags.ContainsStringData);
			}
			else
			{
				return InternalAdd(System.Text.Encoding.Unicode.GetBytes(data), RecordFlags.ContainsStringData);
			}
		}
		
		public byte[] GetData(RecordId recordNumber)
		{
			uint flags;
			return InternalGetData(recordNumber, out flags);
		}
		
		public string GetDataAsString(RecordId recordNumber)
		{
			uint flags;
			byte[] buffer = InternalGetData(recordNumber, out flags);
			if ((buffer != null) && RecordFlags.GetBit(flags, RecordFlags.ContainsStringData))
			{
				return System.Text.Encoding.Unicode.GetString(buffer);
			}
			else
			{
				return null;
			}
		}
		
		private byte[] InternalGetData(RecordId recordNumber, out uint flags)
		{
			flags = 0;
			if (recordNumber == NoRecord)
			{
				return null;
			}
			Connect();
			try
			{
				if (recordNumber > _header.RecordCount)
				{
					return null;
				}
				Record buffer = ReadRecord(recordNumber, true);
				if (RecordFlags.GetBit(buffer.Flags, RecordFlags.Deleted))
				{
					return null;
				}
				flags = buffer.Flags;
				return buffer.Data;
			}
			finally
			{
				Disconnect();
			}
		}
		
		public bool Remove(RecordId recordNumber)
		{
			if (recordNumber == NoRecord)
			{
				return false;
			}
			Connect();
			try
			{
				if (recordNumber > _header.RecordCount)
				{
					return false;
				}
				Record buffer = ReadRecord(recordNumber, false);
				if (!RecordFlags.GetBit(buffer.Flags, RecordFlags.Deleted))
				{
					RecordFlags.SetBit(ref buffer.Flags, RecordFlags.Deleted, true);
					
					if (!RecordFlags.GetBit(buffer.Flags, RecordFlags.Locked))
					{
						// remove record from the data chain
						InternalRemoveFromChain(ref buffer, ref _header.FirstDataRecord, ref _header.LastDataRecord);
						
						// insert record to the free chain
						InternalAddToChain(ref buffer, ref _header.FirstFreeRecord, ref _header.LastFreeRecord);
					}
					
					WriteRecord(ref buffer, false);
					return true;
				}
				else
				{
					return false;
				}
			}
			finally
			{
				Disconnect();
			}
		}
		
		private void InternalRemoveFromChain(ref Record item, ref RecordId head, ref RecordId tail)
		{
			Record previous = ReadRecord(item.PreviousRecordNumber, false);
			Record next = ReadRecord(item.NextRecordNumber, false);
			
			previous.NextRecordNumber = item.NextRecordNumber;
			next.PreviousRecordNumber = item.PreviousRecordNumber;
			
			WriteRecord(ref previous, false);
			WriteRecord(ref next, false);
			
			if (head == item.RecordNumber)
			{
				head = item.NextRecordNumber;
			}
			if (tail == item.RecordNumber)
			{
				tail = item.PreviousRecordNumber;
			}
		}
		
		private void InternalAddToChain(ref Record item, ref RecordId head, ref RecordId tail)
		{
			if (tail == NoRecord)
			{
				item.PreviousRecordNumber = NoRecord;
				item.NextRecordNumber = NoRecord;
				head = item.RecordNumber;
				tail = item.RecordNumber;
			}
			else
			{
				Record previous = ReadRecord(tail, false);
				
				previous.NextRecordNumber = item.RecordNumber;
				item.PreviousRecordNumber = previous.RecordNumber;
				
				WriteRecord(ref previous, false);
				
				item.NextRecordNumber = NoRecord;
				
				tail = item.RecordNumber;
			}
		}
		
		private RecordId InternalAdd(byte[] data, uint flags)
		{
			if ((data != null) && (data.Length > _maxDataSize))
			{
				throw new ArgumentOutOfRangeException();
			}
			
			Connect();
			try
			{
				Record item = new Record();
				if (data == null)
				{
					item.DataSize = 0;
				}
				else
				{
					item.DataSize = (uint)data.Length;
					item.Data = data;
				}
				item.Flags = flags;
				if (_header.LastFreeRecord == NoRecord)
				{
					item.RecordNumber = 1 + _header.RecordCount;
					_header.RecordCount++;
				}
				else
				{
					Record freeItem = ReadRecord(_header.FirstFreeRecord, false);
					InternalRemoveFromChain(ref freeItem, ref _header.FirstFreeRecord, ref _header.LastFreeRecord);
					item.RecordNumber = freeItem.RecordNumber;
				}
				InternalAddToChain(ref item, ref _header.FirstDataRecord, ref _header.LastDataRecord);
			
				WriteRecord(ref item, true);
				
				item.Data = null;
				return item.RecordNumber;
			}
			finally
			{
				Disconnect();
			}
		}
		
		#region Stream I/O
		private void Read(byte[] buffer, uint count)
		{
			int toDo = (int)count;
			int offset = 0;
			do
			{
				int done = _stream.Read(buffer, offset, toDo);
				toDo -= done;
				offset += done;
			}
			while (toDo > 0);
		}
		
		private byte ReadByte()
		{
			return (byte)_stream.ReadByte();
		}
		
		private void Read(out byte buffer)
		{
			buffer = ReadByte();
		}
		
		private void Read(out uint buffer)
		{
			buffer = 0;
			int shift = 0;
			for(int i = 0; i < sizeof(uint); i++, shift += 8)
			{
				buffer = buffer | (((uint)ReadByte()) << shift);
			}
		}
		
		private void Read(out ulong buffer)
		{
			uint lo;
			Read(out lo);
			uint hi;
			Read(out hi);
			buffer = (((ulong)hi) << 32) | lo;
		}
		
		private void Write(byte[] buffer, uint count)
		{
			_stream.Write(buffer, 0, (int)count);
		}
		
		private void Write(byte value)
		{
			_stream.WriteByte(value);
		}
		
		private void Write(uint value)
		{
			for(int i = 0; i < sizeof(uint); i++)
			{
				_stream.WriteByte((byte)(value & 0xFF));
				value = value >> 8;
			}
		}
		
		private void Write(ulong value)
		{
			Write((uint)(value & 0xFFFFFFFF));
			Write((uint)(value >> 32));
		}
		#endregion
		
		private void WriteHeader()
		{
			SeekToHeader();
			Write(_header.Signature);
			Write(_header.RecordCount);
			Write(_header.FirstDataRecord);
			Write(_header.LastDataRecord);
			Write(_header.FirstFreeRecord);
			Write(_header.LastFreeRecord);
		}

		private int _connectCount = 0;
		private void Connect()
		{
			if (_connectCount > 0)
			{
				_connectCount++;
				return;
			}
			_stream = new System.IO.FileStream(_fileName, System.IO.FileMode.OpenOrCreate, System.IO.FileAccess.ReadWrite);
			try
			{
				if (_stream.Length == 0)
				{
					_header = new Header();
					_header.Signature = HeaderSignature;
				}
				else
				{
					Read(out _header.Signature);
					if (_header.Signature != HeaderSignature)
					{
						throw new NotSupportedException();
					}
					Read(out _header.RecordCount);
					Read(out _header.FirstDataRecord);
					Read(out _header.LastDataRecord);
					Read(out _header.FirstFreeRecord);
					Read(out _header.LastFreeRecord);
				}
			}
			catch(Exception)
			{
				_stream.Dispose();
				_stream = null;
				throw;
			}
			_connectCount++;
		}
		
		private void Disconnect()
		{
			_connectCount--;
			if (_connectCount <= 0)
			{
				if (_stream != null)
				{
					WriteHeader();
					_stream.Dispose();
					_stream = null;
				}
			}
		}
	}
}
