using System;

namespace System.IO
{
	/// <summary>
	/// This class captures WriteLine(string value) calls and converts leading spaces to tabs. Use TabSize property for configuration.
	/// </summary>
	public class TabbedTextWriter: StreamWriter
	{
		public class Options
		{
			private uint pInput = 8;
			public uint Input
			{
				get
				{
					return pInput;
				}
				set
				{
					if (value == 0)
					{
						throw new ArgumentOutOfRangeException();
					}
					else
					{
						pInput = value;
					}
				}
			}

			private uint pOutput = 8;
			public uint Output
			{
				get
				{
					return pOutput;
				}
				set
				{
					if (value == 0)
					{
						throw new ArgumentOutOfRangeException();
					}
					else
					{
						pOutput = value;
					}
				}
			}

			public enum Conversion
			{
				None,
				SpacesToTabs,
				TabsToSpaces
			}

			private Conversion pOutputMode = Conversion.SpacesToTabs;
			public Conversion OutputMode
			{
				get
				{
					return pOutputMode;
				}
				set
				{
					pOutputMode = value;
				}
			}
		}
		private Options pTabSize = new Options();
		public Options TabSize
		{
			get
			{
				return pTabSize;
			}
		}

		private bool pModified = false;
		public bool Modified
		{
			get
			{
				return pModified;
			}
			set
			{
				pModified = value;
			}
		}

		public TabbedTextWriter(string path, bool append, System.Text.Encoding encoding)
			:base(path, append, encoding)
		{
		}

		public TabbedTextWriter(string path): base(path)
		{
		}

		#region Workers
		private static string GetTabsAndSpaces(int countOfSpaces, uint TabSize)
		{
			string result = "";
			while(countOfSpaces >= TabSize)
			{
				result = result + "\t";
				countOfSpaces -= (int)TabSize;
			}
			while (countOfSpaces > 0)
			{
				result = result + " ";
				countOfSpaces--;
			}
			return result;
		}

		private static string GetSpaces(int countOfSpaces)
		{
			string result = "";
			while (countOfSpaces > 0)
			{
				result = result + " ";
				countOfSpaces--;
			}
			return result;
		}
		#endregion

		private static bool Convert(ref string line, uint inputTabSize, uint outputTabSize, Options.Conversion mode)
		{
			// TODO: convert also non-leading spaces at white-space positions.
			// white-space position would be defined by a virtual parsing method
			if ((line == null) || (mode == Options.Conversion.None)) return false;

			int toDelete = 0;
			int indent = 0;
			while (toDelete < line.Length)
			{
				char ch = line[toDelete];

				if (ch == ' ')
				{
					toDelete++;
					indent++;
				}
				else if(ch == '\t')
				{
					toDelete++;
					indent++;
					while ((indent % inputTabSize) != 0)
					{
						indent++;
					}
				}
				else
				{
					break;
				}
			}

			if (indent > 0)
			{
				if (toDelete == line.Length)
				{
					// blank line
					line = "";
					return true;
				}

				string prefix = "";
				switch (mode)
				{
					case Options.Conversion.SpacesToTabs:
						prefix = GetTabsAndSpaces(indent, outputTabSize);
						break;

					case Options.Conversion.TabsToSpaces:
						prefix = GetSpaces(indent);
						break;
				}
				string result = prefix + line.Substring(toDelete);
				if (result != line)
				{
					line = result;
					return true;
				}
			}
			return false;
		}

		public string Convert(string value)
		{
			string line = value;
			Convert(ref line, TabSize.Input, TabSize.Output, TabSize.OutputMode);
			return line;
		}

		public override void WriteLine(string value)
		{
			string line = value;
			if (Convert(ref line, TabSize.Input, TabSize.Output, TabSize.OutputMode))
			{
				pModified = true;
			}
			base.WriteLine(line);
		}
	}
}
