using System;

namespace System.Text
{
	/// <summary>
	/// Description of FileStream.
	/// </summary>
	public class FileStream
	{
		public static Encoding GetEncoding(string fileName)
		{
			return GetEncoding(fileName, null);
		}

		public static Encoding GetEncoding(string fileName, int defaultCodePage)
		{
			Encoding defaultEncoding;
			try
			{
				defaultEncoding = Encoding.GetEncoding(defaultCodePage);
			}
			catch(Exception)
			{
				defaultEncoding = Encoding.Default;
			}
			return GetEncoding(fileName, defaultEncoding);
		}

		public static Encoding GetEncoding(System.IO.Stream stream)
		{
			return GetEncoding(stream, null);
		}
		
		public static Encoding GetEncoding(System.IO.Stream stream, Encoding defaultEncoding)
		{
			if (defaultEncoding == null)
			{
				defaultEncoding = Encoding.Default;
			}
			if (stream == null)
			{
				return defaultEncoding;
			}
			
			long oldPosition = stream.Position;
			try
			{
				stream.Position = 0;
				try
				{
					int magic1 = stream.ReadByte();
					int magic2 = stream.ReadByte();
					int magic3 = stream.ReadByte();
					int magic4 = stream.ReadByte();

					if ((magic1 == 0xFF) && (magic2 == 0xFE))
					{
						if ((magic3 == 0x00) && (magic4 == 0x00))
						{
							return Encoding.UTF32;
						}
						else
						{
							return Encoding.Unicode;
						}
					}
					
					if ((magic1 == 0xFE) && (magic2 == 0xFF))
					{
						return Encoding.BigEndianUnicode;
					}
					
					if ((magic1 == 0xEF) && (magic2 == 0xBB) && (magic3 == 0xBF))
					{
						return Encoding.UTF8;
					}

					// TODO: check other encodings using Encoding.GetPreamble()
					return defaultEncoding;
				}
				catch(Exception)
				{
					// TODO: at this point the file could not be opened
					return defaultEncoding;
				}
			}
			finally
			{
				stream.Position = oldPosition;
			}
		}
		
		private static System.IO.Stream OpenReaderStream(string fileName)
		{
			return new System.IO.FileStream(fileName, System.IO.FileMode.Open, System.IO.FileAccess.Read, System.IO.FileShare.ReadWrite);
		}
		
		public static Encoding GetEncoding(string fileName, Encoding defaultEncoding)
		{
			if (defaultEncoding == null)
			{
				defaultEncoding = Encoding.Default;
			}
			if (!System.IO.File.Exists(fileName))
			{
				return defaultEncoding;
			}

			System.IO.Stream stream = OpenReaderStream(fileName);
			try
			{
				return GetEncoding(stream, defaultEncoding);
			}
			finally
			{
				stream.Close();
			}
		}

		private class Reader: System.IO.StreamReader
		{
			public Reader(System.IO.Stream stream, System.Text.Encoding encoding): base(stream, encoding)
			{
			}

			/* TODO: is this code really able to eat both unix-style and windows-style newline separators?
			private char _newLinePair = char.MinValue;
			
			public override string ReadLine()
			{
				bool initialized = false;
				System.Text.StringBuilder result = new System.Text.StringBuilder();
				while (true)
				{
					int code = Read();
					if (code == -1)
					{
						if (!initialized)
						{
							return null;
						}
						else
						{
							return result.ToString();
						}
					}
					char c = (char)code;
					switch (c)
					{
						case '\r':
							if (_newLinePair == '\r')
							{
								_newLinePair = char.MinValue;
								break;
							}
							else
							{
								_newLinePair = '\n';
								return result.ToString();
							}
							
						case '\n':
							if (_newLinePair == '\n')
							{
								_newLinePair = char.MinValue;
								break;
							}
							else
							{
								_newLinePair = '\r';
								return result.ToString();
							}
							 
						default:
							_newLinePair = char.MinValue;
							initialized = true;
							result.Append(c);
							break;
					}
				}
			}
			*/
		}
		
		public static System.IO.TextReader CreateReader(string fileName, Encoding defaultEncoding)
		{
			return new Reader(OpenReaderStream(fileName), GetEncoding(fileName, defaultEncoding));
		}

		public static System.IO.TextReader CreateReader(string fileName, int defaultCodePage)
		{
			return new Reader(OpenReaderStream(fileName), GetEncoding(fileName, defaultCodePage));
		}

		public static System.IO.TextReader CreateReader(string fileName)
		{
			return CreateReader(fileName, null);
		}
		
		public static System.IO.TextReader CreateReader(System.IO.Stream stream)
		{
			return new Reader(stream, GetEncoding(stream));
		}
		
		public static System.IO.TextReader CreateReader(System.IO.Stream stream, Encoding defaultEncoding)
		{
			return new Reader(stream, GetEncoding(stream, defaultEncoding));
		}
	}
}
