using System;
using System.Security;
using System.Runtime.InteropServices;
using System.Collections;

namespace System.Windows.Forms
{
	public class IdleTask: IDisposable
	{
		static IdleQueue guard;
		static IdleTask()
		{
			guard = new IdleQueue();
		}
		internal QueueEntry queue;

		#region System.IDisposable interface implementation
		~IdleTask()
		{
			Dispose(false);
		}

		public void Dispose()
		{
			Dispose(true);
			GC.SuppressFinalize(this);
		}

		private void Dispose(bool disposing)
		{
			if(disposing)
			{
				if(queue != null)
				{
					queue.Task = null;
					queue = null;
				}
			}
		}
		#endregion

		public IdleTask()
		{
			guard.Add(this);
		}

		public static IDisposable New(MethodInvoker handler)
		{
			IdleTask result = new IdleTask();
			result.handler = handler;
			return result;
		}

		protected Delegate handler;
		protected object[] args;

		public virtual void Execute()
		{
			handler.DynamicInvoke(args);
		}
	}

	internal class QueueEntry
	{
		public IdleTask Task;
	}

	internal class IdleQueue
	{
		[SuppressUnmanagedCodeSecurity]
		[return: MarshalAs(UnmanagedType.Bool)]
		[DllImport("User32.dll", CharSet = CharSet.Auto, SetLastError = true)]
		private static extern bool PeekMessage(out Message message, IntPtr handle, uint filterMin, uint filterMax, uint flags);		

		[StructLayout(LayoutKind.Sequential)]
		public struct Message 
		{
			public IntPtr handle;
			public uint msg;
			public IntPtr wParam;
			public IntPtr lParam;
			public uint time;
			public System.Drawing.Point p;
		}		
		
		private Queue tasks;
		private bool active = false;

		private void Activate()
		{
			if(!active)
			{
				active = true;
				Application.Idle += new EventHandler(this.Execute);
			}
		}

		private void Deactivate()
		{
			if(active)
			{
				active = false;
				Application.Idle -= new EventHandler(this.Execute);
			}
		}

		public IdleQueue()
		{
			tasks = new Queue();
		}

		public void Add(IdleTask task)
		{
			QueueEntry entry = new QueueEntry();
			entry.Task = task;
			task.queue = entry;
			tasks.Enqueue(entry);
			Activate();
		}

		private bool IsSomeMessageInTheQueue()
		{
			Message msg;
			return PeekMessage(out msg, IntPtr.Zero, 0, 0, 0);
		}
		
		public void Execute(object sender, System.EventArgs e)
		{
			QueueEntry entry;
		loop:
			
			if (tasks.Count == 0)
			{
				// this must be recursive call
				return;
			}
			
			entry = (QueueEntry)tasks.Dequeue();
			if(entry.Task != null)
			{
				entry.Task.queue = null;
				entry.Task.Execute();
			}
			if(tasks.Count == 0)
			{
				Deactivate();
			}
			else
			{
				if (!IsSomeMessageInTheQueue())
				{
					// take another task
					System.Threading.Thread.Sleep(0);
					goto loop;
				}
				else
				{
					return;
				}
			}
		}
	}
}
